from marshmallow import Schema, RAISE, fields

from argrelay.misc_helper.TypeDesc import TypeDesc
from argrelay.schema_config_interp.InitControlSchema import init_control_desc
from argrelay.schema_config_interp.SearchControlSchema import search_control_desc

function_search_control_ = "function_search_control"

function_init_control_ = "function_init_control"


class FuncArgsInterpConfigSchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    function_search_control = fields.Nested(
        search_control_desc.dict_schema,
        required = True,
    )

    function_init_control = fields.Nested(
        init_control_desc.dict_schema,
        required = True,
    )


func_args_interp_config_example = {
    function_search_control_: search_control_desc.dict_example,
    function_init_control_: init_control_desc.dict_example,
}

func_args_interp_config_desc = TypeDesc(
    dict_schema = FuncArgsInterpConfigSchema(),
    ref_name = FuncArgsInterpConfigSchema.__name__,
    dict_example = func_args_interp_config_example,
    default_file_path = "",
)



"""
"""
import os
from contextlib import contextmanager
from inspect import getframeinfo, currentframe
from pathlib import PurePath

test_data_ = "test_data"

# Enable/disable breakpoints programmatically:
__is_breakpoint_enabled = False


def enable_breakpoint():
    global __is_breakpoint_enabled
    __is_breakpoint_enabled = True


def disable_breakpoint():
    global __is_breakpoint_enabled
    __is_breakpoint_enabled = False


def is_breakpoint_enabled() -> bool:
    return __is_breakpoint_enabled


def line_no() -> int:
    """
    Get source line in the caller frame
    """
    return getframeinfo(currentframe().f_back).lineno


def parse_line_and_cpos(test_line: str) -> (str, int):
    """
    Translate test line with pipe as cursor place into command line string and cursor char position (cpos)
    """
    assert test_line.count('|') == 1
    cursor_cpos = test_line.index('|')
    command_line = test_line.replace('|', "")
    return command_line, cursor_cpos


@contextmanager
def change_to_known_repo_path(path_from_repo_root = "./tests"):
    """
    This function changes to known path within repo root.

    This allows any other code relying on the file access within the repo use reliable relative paths.
    """

    old_pwd = os.getcwd()
    try:
        # When IDE runs, CWD is somewhere under "tests".
        # Keep climbing up until directory becomes "tests":
        if "tests" in PurePath(os.getcwd()).parts:
            while os.path.basename(os.getcwd()) != "tests":
                os.chdir("..")

        # When `tox` runs, CWD = [repo root]:
        if os.path.basename(os.getcwd()) != "tests":
            os.chdir("tests")

        # Now, when same baseline is set regardless `tox` or IDE, change to repo root:
        os.chdir("..")
        # Desired path:
        os.chdir(path_from_repo_root)
        yield
    finally:
        os.chdir(old_pwd)



from __future__ import annotations

from dataclasses import dataclass, field

from argrelay.enum_desc.TermColor import TermColor
from argrelay.misc_helper import eprint
from argrelay.runtime_context.EnvelopeContainer import EnvelopeContainer


# TODO: Combine InterpResult and InvocationInput - only one need to exits
@dataclass
class InterpResult:
    """
    See :class:`InterpResultSchema`
    """

    all_tokens: list[str] = field()
    """
    Copy from `ParsedContext.all_tokens` - command line args.
    """

    consumed_tokens: list[int] = field()
    """
    Copy from `InterpContext.consumed_tokens` -
    indexes into `all_tokens` pointing to tokens consumed during interpretation.
    """

    envelope_containers: list[EnvelopeContainer] = field()

    def describe_data(self):
        eprint()

        for i in range(len(self.all_tokens)):
            if i in self.consumed_tokens:
                eprint(f"{TermColor.BRIGHT_BLUE.value}{self.all_tokens[i]}{TermColor.RESET.value}", end = " ")
            else:
                eprint(f"{TermColor.DARK_MAGENTA.value}{self.all_tokens[i]}{TermColor.RESET.value}", end = " ")

        EnvelopeContainer.describe_data(self.envelope_containers)



from argrelay.enum_desc.CompType import CompType

from argrelay.handler_request.AbstractServerRequestHandler import AbstractServerRequestHandler
from argrelay.misc_helper.ElapsedTime import ElapsedTime
from argrelay.relay_server.LocalServer import LocalServer
from argrelay.runtime_context.InputContext import InputContext
from argrelay.schema_response.InterpResultSchema import (
    interp_result_desc,
    envelope_containers_,
    all_tokens_,
    consumed_tokens_,
)


class DescribeLineArgsServerRequestHandler(AbstractServerRequestHandler):

    def __init__(
        self,
        local_server: LocalServer,
    ):
        super().__init__(
            local_server = local_server,
        )

    def handle_request(self, input_ctx: InputContext) -> dict:
        assert input_ctx.comp_type == CompType.DescribeArgs

        self.interpret_command(self.local_server, input_ctx)
        ElapsedTime.measure("after_interpret_command")

        response_dict = interp_result_desc.dict_schema.dump({
            all_tokens_: self.interp_ctx.parsed_ctx.all_tokens,
            consumed_tokens_: self.interp_ctx.consumed_tokens,
            envelope_containers_: self.interp_ctx.envelope_containers,
        })
        return response_dict



from marshmallow import Schema, RAISE, fields, pre_dump, post_load

from argrelay.misc_helper.TypeDesc import TypeDesc
from argrelay.schema_response.EnvelopeContainerSchema import envelope_container_desc
from argrelay.schema_response.InterpResult import InterpResult

"""
Schema for the result of interpretation taken from :class:`InterpContext`
"""

all_tokens_ = "all_tokens"
consumed_tokens_ = "consumed_tokens"
envelope_containers_ = "envelope_containers"


class InterpResultSchema(Schema):
    class Meta:
        unknown = RAISE
        ordered = True

    all_tokens = fields.List(
        fields.String(),
        required = True,
    )

    consumed_tokens = fields.List(
        fields.Integer(),
        required = True,
    )

    envelope_containers = fields.List(
        fields.Nested(envelope_container_desc.dict_schema),
        required = True,
    )

    @pre_dump
    def make_dict(self, input_object: InterpResult, **kwargs):
        if isinstance(input_object, InterpResult):
            return {
                all_tokens_: input_object.all_tokens,
                consumed_tokens_: input_object.consumed_tokens,
                envelope_containers_: input_object.envelope_containers,
            }
        else:
            # Assuming it is as dict:
            return input_object
        pass

    @post_load
    def make_object(self, input_dict, **kwargs):
        return InterpResult(
            all_tokens = input_dict[all_tokens_],
            consumed_tokens = input_dict[consumed_tokens_],
            envelope_containers = input_dict[envelope_containers_],
        )


interp_result_desc = TypeDesc(
    dict_schema = InterpResultSchema(),
    ref_name = InterpResultSchema.__name__,
    dict_example = {
        all_tokens_: [
            "some_command",
            "unrecognized_token",
            "goto",
            "host",
            "prod",
        ],
        consumed_tokens_: [
            0,
            2,
            3,
            4,
        ],
        envelope_containers_: [
            envelope_container_desc.dict_example,
        ],
    },
    default_file_path = "",
)



from marshmallow import Schema, fields, RAISE, post_load

from argrelay.misc_helper.TypeDesc import TypeDesc
from argrelay.mongo_data.MongoConfig import MongoConfig
from argrelay.schema_config_core_server.MongoClientConfigSchema import mongo_client_config_desc
from argrelay.schema_config_core_server.MongoServerConfigSchema import mongo_server_config_desc

use_mongomock_only_ = "use_mongomock_only"
mongo_client_ = "mongo_client"
mongo_server_ = "mongo_server"


class MongoConfigSchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    use_mongomock_only = fields.Boolean()
    """
    It might be the case that (test) `mongomock` lib actually provides necessary functionality and meet
    non-function requirements for many (prod) workloads without need to deploy and administer MongoDB.
    https://github.com/mongomock/mongomock
    """

    mongo_client = fields.Nested(mongo_client_config_desc.dict_schema)

    mongo_server = fields.Nested(mongo_server_config_desc.dict_schema)

    @post_load
    def make_object(self, input_dict, **kwargs):
        return MongoConfig(
            use_mongomock_only = input_dict[use_mongomock_only_],
            mongo_client = input_dict[mongo_client_],
            mongo_server = input_dict[mongo_server_],
        )


mongo_config_desc = TypeDesc(
    dict_schema = MongoConfigSchema(),
    ref_name = MongoConfigSchema.__name__,
    dict_example = {
        use_mongomock_only_: True,
        mongo_client_: mongo_client_config_desc.dict_example,
        mongo_server_: mongo_server_config_desc.dict_example,
    },
    default_file_path = "",
)



from argrelay.enum_desc.CompType import CompType
from argrelay.enum_desc.RunMode import RunMode

from argrelay.relay_client import AbstractClientCommand
from argrelay.runtime_context.InputContext import InputContext
from argrelay.server_spec.const_int import (
    DESCRIBE_LINE_ARGS_PATH,
    RELAY_LINE_ARGS_PATH,
    PROPOSE_ARG_VALUES_PATH,
)


class AbstractClientCommandFactory:

    def create_command_by_server_path(self, server_path: str) -> AbstractClientCommand:
        pass

    def create_command(self, input_ctx: InputContext):
        server_path = self.select_command(input_ctx)
        return self.create_command_by_server_path(server_path)

    @staticmethod
    def select_command(input_ctx: InputContext) -> str:
        if input_ctx.run_mode == RunMode.InvocationMode:
            if input_ctx.comp_type == CompType.DescribeArgs:
                return DESCRIBE_LINE_ARGS_PATH
            else:
                assert input_ctx.comp_type == CompType.InvokeAction
                return RELAY_LINE_ARGS_PATH
        else:
            assert input_ctx.run_mode == RunMode.CompletionMode
            return PROPOSE_ARG_VALUES_PATH



from argrelay.misc_helper.AbstractPlugin import AbstractPlugin
from argrelay.plugin_interp.AbstractInterp import AbstractInterp
from argrelay.runtime_context.InterpContext import InterpContext


class AbstractInterpFactory(AbstractPlugin):

    def create_interp(
        self,
        interp_ctx: InterpContext,
    ) -> AbstractInterp:
        pass



from marshmallow import Schema, fields, RAISE, post_load, pre_dump

from argrelay.enum_desc.ArgSource import ArgSource
from argrelay.misc_helper import ensure_value_is_enum
from argrelay.misc_helper.TypeDesc import TypeDesc
from argrelay.runtime_data.AssignedValue import AssignedValue

arg_value_ = "arg_value"
arg_source_ = "arg_source"


class AssignedValueSchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    arg_value = fields.String(
        required = True,
    )

    arg_source = fields.Enum(
        ArgSource,
        by_value = False,
        required = True,
    )

    @pre_dump
    def make_dict(self, input_object: AssignedValue, **kwargs):
        # TODO: figure out to populate all automatically and reduce duplication - this is error-prone:
        if isinstance(input_object, AssignedValue):
            return {
                arg_value_: input_object.arg_value,
                arg_source_: ensure_value_is_enum(input_object.arg_source, ArgSource),
            }
        else:
            # Assuming it is as dict:
            return {
                arg_value_: input_object[arg_value_],
                arg_source_: ensure_value_is_enum(input_object[arg_source_], ArgSource),
            }

    @post_load
    def make_object(self, input_dict, **kwargs):
        return AssignedValue(
            arg_value = input_dict[arg_value_],
            arg_source = ensure_value_is_enum(input_dict[arg_source_], ArgSource),
        )


assigned_value_desc = TypeDesc(
    dict_schema = AssignedValueSchema(),
    ref_name = AssignedValueSchema.__name__,
    dict_example = {
        arg_value_: "value_1",
        arg_source_: ArgSource.ImplicitValue.name,
    },
    default_file_path = "",
)



from argrelay.enum_desc.RunMode import RunMode

from argrelay.relay_server.LocalServer import LocalServer
from argrelay.runtime_context.InputContext import InputContext
from argrelay.runtime_context.InterpContext import InterpContext
from argrelay.runtime_context.ParsedContext import ParsedContext
from argrelay.runtime_context.RequestContext import RequestContext


class AbstractServerRequestHandler:

    def __init__(
        self,
        local_server: LocalServer,
    ):
        self.local_server: LocalServer = local_server
        self.interp_ctx: InterpContext

    def handle_request(self, input_ctx: InputContext) -> dict:
        raise NotImplementedError

    @staticmethod
    def create_input_ctx(request_ctx: RequestContext, run_mode: RunMode):
        input_ctx = InputContext.from_request_context(
            request_ctx,
            run_mode = run_mode,
            comp_key = str(0),
        )
        return input_ctx

    def interpret_command(self, local_server: LocalServer, input_ctx: InputContext):
        parsed_ctx = ParsedContext.from_instance(input_ctx)
        # TODO: Split server_config and static_data (both top level, not config including data):
        self.interp_ctx = InterpContext(
            parsed_ctx = parsed_ctx,
            interp_factories = local_server.server_config.interp_factories,
            action_delegators = local_server.server_config.action_delegators,
            query_engine = local_server.get_query_engine(),
            help_hint_cache = local_server.help_hint_cache,
        )
        self.interp_ctx.interpret_command(local_server.server_config.static_data.first_interp_factory_id)
        self.interp_ctx.print_debug()



from __future__ import annotations

import subprocess
from argrelay.custom_integ.value_constants import goto_repo_func_, desc_commit_func_
from argrelay.misc_helper import eprint
from argrelay.plugin_delegator.AbstractDelegator import AbstractDelegator, get_data_envelopes
from argrelay.plugin_delegator.InvocationInput import InvocationInput
from argrelay.relay_server.LocalServer import LocalServer
from argrelay.runtime_context.InterpContext import InterpContext, function_envelope_ipos_
from argrelay.schema_config_interp.DataEnvelopeSchema import envelope_payload_, envelope_id_, instance_data_
from argrelay.schema_config_interp.FunctionEnvelopeInstanceDataSchema import delegator_plugin_instance_id_

repo_envelope_ipos_ = 1


class GitRepoDelegator(AbstractDelegator):
    """
    Implements FS_67_16_61_97 git_plugin.
    """

    def __init__(
        self,
        plugin_instance_id: str,
        config_dict: dict,
    ):
        super().__init__(
            plugin_instance_id,
            config_dict,
        )

    def run_invoke_control(
        self,
        interp_ctx: InterpContext,
        local_server: LocalServer,
    ) -> InvocationInput:

        assert interp_ctx.is_funct_found(), "the (first) function envelope must be found"

        # The first envelope (`DataEnvelopeSchema`) is assumed to be of
        # `ReservedEnvelopeClass.ClassFunction` with `FunctionEnvelopeInstanceDataSchema` for its `instance_data`:
        function_envelope = interp_ctx.envelope_containers[function_envelope_ipos_]
        delegator_plugin_instance_id = function_envelope.data_envelope[instance_data_][delegator_plugin_instance_id_]
        invocation_input = InvocationInput(
            all_tokens = interp_ctx.parsed_ctx.all_tokens,
            consumed_tokens = interp_ctx.consumed_tokens,
            delegator_plugin_entry = local_server.server_config.plugin_dict[delegator_plugin_instance_id],
            data_envelopes = get_data_envelopes(interp_ctx),
            custom_plugin_data = {},
        )
        return invocation_input

    @staticmethod
    def invoke_action(invocation_input: InvocationInput):
        if invocation_input.data_envelopes[function_envelope_ipos_][envelope_id_] == goto_repo_func_:
            repo_envelope = invocation_input.data_envelopes[repo_envelope_ipos_]
            repo_root_abs_path = repo_envelope[envelope_payload_][repo_root_abs_path_]
            eprint(f"INFO: starting subshell in: {repo_root_abs_path}")
            # List Git repo dir:
            subproc = subprocess.run(
                [
                    "bash",
                    "-l",
                ],
                cwd = repo_root_abs_path
            )
            ret_code = subproc.returncode
            if ret_code != 0:
                raise RuntimeError
        if invocation_input.data_envelopes[function_envelope_ipos_][envelope_id_] == desc_commit_func_:
            raise RuntimeError("not implemented")


repo_root_abs_path_: str = "repo_root_abs_path"



from argrelay.enum_desc.CompType import CompType

from argrelay.handler_request.AbstractServerRequestHandler import AbstractServerRequestHandler
from argrelay.misc_helper.ElapsedTime import ElapsedTime
from argrelay.plugin_delegator.AbstractDelegator import AbstractDelegator
from argrelay.plugin_delegator.ErrorDelegator import ErrorDelegator
from argrelay.plugin_delegator.ErrorDelegatorCustomDataSchema import (
    error_message_,
    error_delegator_custom_data_desc,
    error_code_,
)
from argrelay.plugin_delegator.InvocationInput import InvocationInput
from argrelay.relay_server.LocalServer import LocalServer
from argrelay.runtime_context.InputContext import InputContext
from argrelay.runtime_context.InterpContext import function_envelope_ipos_
from argrelay.schema_config_interp.DataEnvelopeSchema import instance_data_
from argrelay.schema_config_interp.FunctionEnvelopeInstanceDataSchema import delegator_plugin_instance_id_
from argrelay.schema_response.InvocationInputSchema import invocation_input_desc


class RelayLineArgsServerRequestHandler(AbstractServerRequestHandler):

    def __init__(
        self,
        local_server: LocalServer,
    ):
        super().__init__(
            local_server = local_server,
        )

    def handle_request(self, input_ctx: InputContext) -> dict:
        assert input_ctx.comp_type == CompType.InvokeAction

        self.interpret_command(self.local_server, input_ctx)
        ElapsedTime.measure("after_interpret_command")
        is_error = False
        error_message = ""
        error_code = 0

        # The first envelope (`DataEnvelopeSchema`) is assumed to be of
        # `ReservedEnvelopeClass.ClassFunction` with `FunctionEnvelopeInstanceDataSchema` for its `instance_data`:
        if self.interp_ctx.is_funct_found():
            delegator_plugin_instance_id = self.interp_ctx.envelope_containers[function_envelope_ipos_].data_envelope[
                instance_data_
            ][
                delegator_plugin_instance_id_
            ]
        else:
            is_error = True
            error_message = "ERROR: Function is not selected, try help, or press Tab to complete selection."
            error_code = 1
            delegator_plugin_instance_id = ErrorDelegator.__name__

        delegator_plugin: AbstractDelegator = self.local_server.server_config.action_delegators[
            delegator_plugin_instance_id
        ]
        invocation_input: InvocationInput = delegator_plugin.run_invoke_control(
            self.interp_ctx,
            self.local_server,
        )

        if is_error:
            invocation_input.custom_plugin_data = {
                error_message_: error_message,
                error_code_: error_code,
            }
            error_delegator_custom_data_desc.validate_dict(invocation_input.custom_plugin_data)

        response_dict = invocation_input_desc.dict_schema.dump(invocation_input)
        return response_dict



from enum import Enum, auto


class GlobalArgType(Enum):
    FunctionCategory = auto()
    """
    A way to separate functions in different "buckets".

    For example:
    *   "internal" provided by `argrelay` for various built-in and support functions
    *   "external" provided by plugins for domain-specific
    """

    ActionType = auto()
    """
    Specifies what function does. For example: "goto", "list", ...
    """

    ObjectSelector = auto()
    """
    Specifies what kind of objects a function works with. For example: "host", "service", ...
    """



from dataclasses import dataclass, field

from argrelay.enum_desc.CompType import CompType


@dataclass(frozen = True)
class RequestContext:
    """
    Part of :class:`InputContext` which is used in requests from client to server.
    """

    command_line: str = field()
    cursor_cpos: int = field()
    comp_type: CompType = field()
    is_debug_enabled: bool = field()

    def __post_init__(self):
        assert 0 <= self.cursor_cpos <= len(self.command_line)



from argrelay.plugin_interp.AbstractInterpFactory import AbstractInterpFactory
from argrelay.plugin_interp.FuncArgsInterp import FuncArgsInterp
from argrelay.runtime_context.InterpContext import InterpContext
from argrelay.runtime_context.SearchControl import SearchControl
from argrelay.schema_config_interp.FuncArgsInterpConfigSchema import (
    func_args_interp_config_desc,
    function_search_control_,
)
from argrelay.schema_config_interp.SearchControlSchema import keys_to_types_list_


class FuncArgsInterpFactory(AbstractInterpFactory):

    def __init__(
        self,
        plugin_instance_id: str,
        config_dict: dict,
    ):
        super().__init__(
            plugin_instance_id,
            config_dict,
        )
        func_args_interp_config_desc.validate_dict(config_dict)

    def create_interp(
        self,
        interp_ctx: InterpContext,
    ) -> FuncArgsInterp:
        raise NotImplementedError

    def validate_loaded_data(
        self,
        static_data: "StaticData",
    ):
        self._validate_function_envelopes_unambiguously_qualified(static_data)

    def _validate_function_envelopes_unambiguously_qualified(
        self,
        static_data: "StaticData",
    ):
        """
        Verify that all arg types listed in `function_search_control` uniquely identify function

        Otherwise, functions which do not have unique "coordinates" with listed arg types will not be invokable.

        Plan:
        *   Find all envelopes of `ReservedEnvelopeClass.ClassFunction`.
        *   Take those of them which have all arg types listed in plugin config.
        *   Make sure all arg values (ordered by arg type) have unique tuple.
        """

        types_list = SearchControl.convert_list_of_ordered_singular_dicts_to_unordered_dict(
            self.config_dict[function_search_control_][keys_to_types_list_]
        ).values()

        all_typed_values_lists = []
        for data_envelope in static_data.data_envelopes:

            all_types_exist = True
            typed_values_list = []
            for type_name in types_list:
                if type_name not in data_envelope:
                    all_types_exist = False
                    break
                else:
                    typed_values_list.append(data_envelope[type_name])

            if all_types_exist:
                assert typed_values_list not in all_typed_values_lists
                all_typed_values_lists.append(typed_values_list)
            else:
                continue



from flasgger import swag_from
from flask import request, Blueprint, Response, abort

from argrelay.enum_desc.RunMode import RunMode
from argrelay.handler_request.AbstractServerRequestHandler import AbstractServerRequestHandler
from argrelay.handler_request.DescribeLineArgsServerRequestHandler import DescribeLineArgsServerRequestHandler
from argrelay.handler_request.ProposeArgValuesServerRequestHandler import ProposeArgValuesServerRequestHandler
from argrelay.handler_request.RelayLineArgsServerRequestHandler import RelayLineArgsServerRequestHandler
from argrelay.misc_helper.ElapsedTime import ElapsedTime
from argrelay.relay_server.LocalServer import LocalServer
from argrelay.schema_request.RequestContextSchema import request_context_desc
from argrelay.schema_response.ArgValuesSchema import arg_values_desc, arg_values_
from argrelay.schema_response.InterpResultSchema import interp_result_desc
from argrelay.schema_response.InvocationInputSchema import invocation_input_desc
from argrelay.server_spec import DescribeLineArgsSpec, ProposeArgValuesSpec, RelayLineArgsSpec
from argrelay.server_spec.const_int import (
    DESCRIBE_LINE_ARGS_PATH,
    PROPOSE_ARG_VALUES_PATH,
    RELAY_LINE_ARGS_PATH,
)


def create_blueprint_api(local_server: LocalServer):
    blueprint_api = Blueprint("blueprint_api", __name__)

    describe_line_args_handler = DescribeLineArgsServerRequestHandler(local_server)
    propose_arg_values_handler = ProposeArgValuesServerRequestHandler(local_server)
    relay_line_args_handler = RelayLineArgsServerRequestHandler(local_server)

    def create_input_ctx(run_mode: RunMode):
        ElapsedTime.clear_measurements()
        ElapsedTime.measure("before_request_payload_load")
        # TODO: Figure out why:
        #       *   requests by `ProposeArgValuesRemoteClientCommand` arrive as dict
        #       *   requests by `AbstractRemoteClientCommand` arrive as str
        request_ctx = None
        if isinstance(request.json, str):
            request_ctx = request_context_desc.dict_schema.loads(request.json)
        if isinstance(request.json, dict):
            request_ctx = request_context_desc.dict_schema.load(request.json)
        input_ctx = AbstractServerRequestHandler.create_input_ctx(request_ctx, run_mode)
        ElapsedTime.measure("after_input_context_creation")
        ElapsedTime.is_debug_enabled = input_ctx.is_debug_enabled
        return input_ctx

    @blueprint_api.route(
        DESCRIBE_LINE_ARGS_PATH,
        methods = ["post"],
    )
    @swag_from(DescribeLineArgsSpec.spec_data)
    def describe_line_args():
        try:
            input_ctx = create_input_ctx(RunMode.InvocationMode)
            response_dict = describe_line_args_handler.handle_request(input_ctx)

            if request.accept_mimetypes["application/json"] or len(request.accept_mimetypes) == 0:
                response_json = interp_result_desc.dict_schema.dumps(response_dict)
                return Response(
                    response_json,
                    mimetype = "application/json",
                )
            else:
                # Not acceptable:
                abort(406)
        finally:
            ElapsedTime.measure("before_sending_response")

    @blueprint_api.route(
        PROPOSE_ARG_VALUES_PATH,
        methods = ["post"],
    )
    @swag_from(ProposeArgValuesSpec.spec_data)
    def propose_arg_values():
        try:
            input_ctx = create_input_ctx(RunMode.CompletionMode)
            response_dict = propose_arg_values_handler.handle_request(input_ctx)

            if request.accept_mimetypes["text/plain"] or len(request.accept_mimetypes) == 0:
                # Sending plain text by default for simplest clients (who may not even specify headers)
                # also serving perf reasons on client side (trivial parsing, no lib required, minimal imports):
                return Response(
                    "\n".join(response_dict[arg_values_]),
                    mimetype = "text/plain",
                )
            elif request.accept_mimetypes["application/json"]:
                # JSON - parsing lib required:
                response_json = arg_values_desc.dict_schema.dumps(response_dict)
                return Response(
                    response_json,
                    mimetype = "application/json",
                )
            else:
                # Not acceptable:
                abort(406)
        finally:
            ElapsedTime.measure("before_sending_response")

    @blueprint_api.route(
        RELAY_LINE_ARGS_PATH,
        methods = ["post"],
    )
    @swag_from(RelayLineArgsSpec.spec_data)
    def relay_line_args():
        try:
            input_ctx = create_input_ctx(RunMode.InvocationMode)
            response_dict = relay_line_args_handler.handle_request(input_ctx)

            if request.accept_mimetypes["application/json"] or len(request.accept_mimetypes) == 0:
                response_json = invocation_input_desc.dict_schema.dumps(response_dict)
                return Response(
                    response_json,
                    mimetype = "application/json",
                )
            else:
                # Not acceptable:
                abort(406)
        finally:
            ElapsedTime.measure("before_sending_response")

    @blueprint_api.teardown_request
    def show_teardown(exception):
        ElapsedTime.print_all_if_debug()

    return blueprint_api



from marshmallow import Schema, RAISE, fields, post_load, pre_dump

from argrelay.misc_helper.TypeDesc import TypeDesc
from argrelay.plugin_delegator.ErrorDelegatorCustomDataSchema import error_delegator_custom_data_desc
from argrelay.plugin_delegator.InvocationInput import InvocationInput
from argrelay.schema_config_interp.DataEnvelopeSchema import data_envelope_desc, mongo_id_
from argrelay.schema_config_plugin.PluginEntrySchema import plugin_entry_desc
from argrelay.schema_response.FilteredDict import FilteredDict

all_tokens_ = "all_tokens"
consumed_tokens_ = "consumed_tokens"
delegator_plugin_entry_ = "delegator_plugin_entry"
data_envelopes_ = "data_envelopes"
custom_plugin_data_ = "custom_plugin_data"


class InvocationInputSchema(Schema):
    class Meta:
        unknown = RAISE
        ordered = True

    all_tokens = fields.List(
        fields.String(),
        required = True,
    )

    consumed_tokens = fields.List(
        fields.Integer(),
        required = True,
    )

    delegator_plugin_entry = fields.Nested(
        plugin_entry_desc.dict_schema,
        required = True,
    )

    data_envelopes = fields.List(
        FilteredDict(
            filtered_keys = [mongo_id_],
            # Some `data_envelope`-s may not be found by search -
            # instead of redirecting invocation to something like `ErrorDelegator` on server side,
            # send `None` items to decide how to handle that on client side.
            allow_none = True,
        ),
        required = True,
    )

    custom_plugin_data = fields.Dict(
        required = False,
    )

    @pre_dump
    def make_dict(self, input_object: InvocationInput, **kwargs):
        if isinstance(input_object, InvocationInput):
            return {
                all_tokens_: input_object.all_tokens,
                consumed_tokens_: input_object.consumed_tokens,
                delegator_plugin_entry_: input_object.delegator_plugin_entry,
                data_envelopes_: input_object.data_envelopes,
                custom_plugin_data_: input_object.custom_plugin_data,
            }
        else:
            # Assuming it is as dict:
            return input_object
        pass

    @post_load
    def make_object(self, input_dict, **kwargs):
        return InvocationInput(
            all_tokens = input_dict[all_tokens_],
            consumed_tokens = input_dict[consumed_tokens_],
            delegator_plugin_entry = input_dict[delegator_plugin_entry_],
            data_envelopes = input_dict[data_envelopes_],
            custom_plugin_data = input_dict[custom_plugin_data_],
        )


_invocation_input_example = {
    all_tokens_: [
        "some_command",
        "unrecognized_token",
        "goto",
        "host",
        "prod",
    ],
    consumed_tokens_: [
        0,
        2,
        3,
        4,
    ],
    delegator_plugin_entry_: plugin_entry_desc.dict_example,
    data_envelopes_: [
        data_envelope_desc.dict_example,
    ],
    custom_plugin_data_: error_delegator_custom_data_desc.dict_example,
}

invocation_input_desc = TypeDesc(
    dict_schema = InvocationInputSchema(),
    ref_name = InvocationInputSchema.__name__,
    dict_example = _invocation_input_example,
    default_file_path = "",
)



from marshmallow import Schema, RAISE, fields

from argrelay.custom_integ.DemoInterpFactory import DemoInterpFactory
from argrelay.misc_helper.TypeDesc import TypeDesc

first_arg_vals_to_next_interp_factory_ids_ = "first_arg_vals_to_next_interp_factory_ids"


class FirstArgInterpFactoryConfigSchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    first_arg_vals_to_next_interp_factory_ids = fields.Dict(
        keys = fields.String(),
        values = fields.String(),
        required = True,
    )


first_arg_interp_factory_config_example = {
    first_arg_vals_to_next_interp_factory_ids_: {
        "some_command": DemoInterpFactory.__name__,
    },
}
first_arg_interp_factory_config_desc = TypeDesc(
    dict_schema = FirstArgInterpFactoryConfigSchema(),
    ref_name = FirstArgInterpFactoryConfigSchema.__name__,
    dict_example = first_arg_interp_factory_config_example,
    default_file_path = "",
)



from dataclasses import dataclass, field

from argrelay.runtime_data.ConnectionConfig import ConnectionConfig


@dataclass(frozen = True)
class ClientConfig:
    use_local_requests: bool = field()
    connection_config: ConnectionConfig = field()



from __future__ import annotations

from argrelay.custom_integ.ServiceArgType import ServiceArgType
from argrelay.custom_integ.ServiceDelegator import ServiceDelegator
from argrelay.custom_integ.ServiceEnvelopeClass import ServiceEnvelopeClass
from argrelay.custom_integ.ServiceLoaderConfigSchema import (
    service_loader_config_desc,
    test_data_ids_to_load_,
)
from argrelay.custom_integ.value_constants import (
    goto_host_funct_,
    goto_service_funct_,
    list_host_func_,
    list_service_func_,
)
from argrelay.enum_desc.GlobalArgType import GlobalArgType
from argrelay.enum_desc.ReservedArgType import ReservedArgType
from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.misc_helper import eprint
from argrelay.plugin_delegator.NoopDelegator import NoopDelegator
from argrelay.plugin_loader.AbstractLoader import AbstractLoader
from argrelay.runtime_data.StaticData import StaticData
from argrelay.schema_config_interp.DataEnvelopeSchema import (
    envelope_payload_,
    envelope_id_,
    instance_data_,
)
from argrelay.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
    search_control_list_,
)
from argrelay.schema_config_interp.SearchControlSchema import keys_to_types_list_, envelope_class_
from argrelay.test_helper import test_data_

cluster_search_control = {
    envelope_class_: ServiceEnvelopeClass.ClassCluster.name,
    keys_to_types_list_: [
        {"code": ServiceArgType.CodeMaturity.name},
        {"stage": ServiceArgType.FlowStage.name},
        {"region": ServiceArgType.GeoRegion.name},
        {"cluster": ServiceArgType.ClusterName.name},
    ],
}

host_search_control = {
    envelope_class_: ServiceEnvelopeClass.ClassHost.name,
    keys_to_types_list_: [
        # ClassCluster:
        {"code": ServiceArgType.CodeMaturity.name},
        {"stage": ServiceArgType.FlowStage.name},
        {"region": ServiceArgType.GeoRegion.name},
        {"cluster": ServiceArgType.ClusterName.name},
        # ClassHost:
        {"host": ServiceArgType.HostName.name},
        # ---
        {"ip": ServiceArgType.IpAddress.name},
    ],
}

service_search_control = {
    envelope_class_: ServiceEnvelopeClass.ClassService.name,
    keys_to_types_list_: [
        # ClassCluster:
        {"code": ServiceArgType.CodeMaturity.name},
        {"stage": ServiceArgType.FlowStage.name},
        {"region": ServiceArgType.GeoRegion.name},
        {"cluster": ServiceArgType.ClusterName.name},
        # ClassService:
        {"service": ServiceArgType.ServiceName.name},
        # ClassHost:
        {"host": ServiceArgType.HostName.name},
        # ---
        {"status": ServiceArgType.LiveStatus.name},
        {"dc": ServiceArgType.DataCenter.name},
        {"ip": ServiceArgType.IpAddress.name},
    ],
}

access_search_control = {
    envelope_class_: ServiceArgType.AccessType.name,
    keys_to_types_list_: [
        {"access": ServiceArgType.AccessType.name},
    ],
}


# noinspection PyPep8Naming
class ServiceLoader(AbstractLoader):

    def __init__(
        self,
        plugin_instance_id: str,
        config_dict: dict,
    ):
        super().__init__(
            plugin_instance_id,
            config_dict,
        )
        service_loader_config_desc.validate_dict(config_dict)

    def update_static_data(
        self,
        static_data: StaticData,
    ) -> StaticData:

        static_data = self.load_data_envelopes(static_data)

        return static_data

    # noinspection PyMethodMayBeStatic
    def load_data_envelopes(
        self,
        static_data: StaticData,
    ) -> StaticData:
        """
        The loader writes samples into `static_data[data_envelopes_]` simply from code (without any data source).
        """

        # Init type keys (if they do not exist):
        for type_name in [enum_item.name for enum_item in ServiceArgType]:
            if type_name not in static_data.known_arg_types:
                static_data.known_arg_types.append(type_name)

        # TODO: This loader overwrites existing object list (it has to patch it instead).
        #       This is fine for now because `ServiceLoader` is used first in config.
        data_envelopes = []

        self.populate_common_functions(data_envelopes)
        self.populate_common_AccessType(data_envelopes)
        self.populate_TD_63_37_05_36_default(data_envelopes)
        self.populate_TD_76_09_29_31_overlapped(data_envelopes)
        self.populate_TD_38_03_48_51_large_generated(data_envelopes)
        self.populate_TD_43_24_76_58_single(data_envelopes)

        self.generate_envelope_id(data_envelopes)

        self.generate_help_hints(data_envelopes)

        static_data.data_envelopes.extend(data_envelopes)

        return static_data

    # noinspection PyMethodMayBeStatic
    def generate_envelope_id(
        self,
        data_envelopes: list[dict],
    ):
        for data_envelope in data_envelopes:
            if envelope_id_ not in data_envelope:
                if data_envelope[ReservedArgType.EnvelopeClass.name] == ServiceEnvelopeClass.ClassHost.name:
                    data_envelope[envelope_id_] = (
                        data_envelope[ServiceArgType.ClusterName.name]
                        + "." +
                        data_envelope[ServiceArgType.HostName.name]
                    )
                if data_envelope[ReservedArgType.EnvelopeClass.name] == ServiceEnvelopeClass.ClassService.name:
                    data_envelope[envelope_id_] = (
                        data_envelope[ServiceArgType.ClusterName.name]
                        + "." +
                        data_envelope[ServiceArgType.HostName.name]
                        + "." +
                        data_envelope[ServiceArgType.ServiceName.name]
                    )

    @staticmethod
    def generate_help_hints(
        data_envelopes: list[dict],
    ):
        """
        Function demos FS_71_87_33_52 help_hint for `ServiceArgType.IpAddress`.

        It simply generates `data_envelope`-s of `ReservedEnvelopeClass.ClassHelp` for
        values of `ServiceArgType.IpAddress` equal to corresponding `ServiceArgType.HostName`.
        """
        help_hint_envelopes = []
        for data_envelope in data_envelopes:
            if data_envelope[ReservedArgType.EnvelopeClass.name] == ServiceEnvelopeClass.ClassHost.name:
                if ServiceArgType.IpAddress.name in data_envelope:
                    help_hint_envelopes.append({
                        f"{ReservedArgType.EnvelopeClass.name}": f"{ReservedEnvelopeClass.ClassHelp.name}",
                        f"{ReservedArgType.ArgType.name}": f"{ServiceArgType.IpAddress.name}",
                        f"{ReservedArgType.ArgValue.name}": f"{data_envelope[ServiceArgType.IpAddress.name]}",
                        f"{ReservedArgType.HelpHint.name}": f"{data_envelope[ServiceArgType.HostName.name]}",
                    })

        data_envelopes.extend(help_hint_envelopes)

    @staticmethod
    def populate_common_functions(
        data_envelopes: list,
    ):
        data_envelopes.extend([

            ############################################################################################################
            # functions

            # TODO: As of now, `ServiceLoader` is configured first (after `GitRepoLoader`), so it works.
            #       But to be robust against re-ordering, this loader should share function names like
            #       "desc", "list", "goto" to accept `GitRepoEnvelopeClass` in addition to `ServiceEnvelopeClass`.
            {
                envelope_id_: goto_host_funct_,
                instance_data_: {
                    delegator_plugin_instance_id_: ServiceDelegator.__name__,
                    search_control_list_: [
                        host_search_control,
                        access_search_control,
                    ],
                },
                ReservedArgType.EnvelopeClass.name: ReservedEnvelopeClass.ClassFunction.name,
                ReservedArgType.HelpHint.name: "Go (log in) to remote host",
                GlobalArgType.FunctionCategory.name: "external",
                GlobalArgType.ActionType.name: "goto",
                GlobalArgType.ObjectSelector.name: "host",
            },
            {
                envelope_id_: goto_service_funct_,
                instance_data_: {
                    delegator_plugin_instance_id_: ServiceDelegator.__name__,
                    search_control_list_: [
                        service_search_control,
                        access_search_control,
                    ],
                },
                ReservedArgType.EnvelopeClass.name: ReservedEnvelopeClass.ClassFunction.name,
                ReservedArgType.HelpHint.name: "Go (log in) to remote host and dir path with specified service",
                GlobalArgType.FunctionCategory.name: "external",
                GlobalArgType.ActionType.name: "goto",
                GlobalArgType.ObjectSelector.name: "service",
            },
            {
                envelope_id_: "desc_host",
                instance_data_: {
                    delegator_plugin_instance_id_: NoopDelegator.__name__,
                    search_control_list_: [
                        host_search_control,
                    ],
                },
                ReservedArgType.EnvelopeClass.name: ReservedEnvelopeClass.ClassFunction.name,
                ReservedArgType.HelpHint.name: "Describe remote host",
                GlobalArgType.FunctionCategory.name: "external",
                GlobalArgType.ActionType.name: "desc",
                GlobalArgType.ObjectSelector.name: "host",
            },
            {
                envelope_id_: "desc_service",
                instance_data_: {
                    delegator_plugin_instance_id_: NoopDelegator.__name__,
                    search_control_list_: [
                        service_search_control,
                    ],
                },
                ReservedArgType.EnvelopeClass.name: ReservedEnvelopeClass.ClassFunction.name,
                ReservedArgType.HelpHint.name: "Describe service instance",
                GlobalArgType.FunctionCategory.name: "external",
                GlobalArgType.ActionType.name: "desc",
                GlobalArgType.ObjectSelector.name: "service",
            },
            {
                envelope_id_: list_host_func_,
                instance_data_: {
                    delegator_plugin_instance_id_: ServiceDelegator.__name__,
                    search_control_list_: [
                        host_search_control,
                    ],
                },
                ReservedArgType.EnvelopeClass.name: ReservedEnvelopeClass.ClassFunction.name,
                ReservedArgType.HelpHint.name: "List remote hosts matching search query",
                GlobalArgType.FunctionCategory.name: "external",
                GlobalArgType.ActionType.name: "list",
                GlobalArgType.ObjectSelector.name: "host",
            },
            {
                envelope_id_: list_service_func_,
                instance_data_: {
                    delegator_plugin_instance_id_: ServiceDelegator.__name__,
                    search_control_list_: [
                        service_search_control,
                    ],
                },
                ReservedArgType.EnvelopeClass.name: ReservedEnvelopeClass.ClassFunction.name,
                ReservedArgType.HelpHint.name: "List service instances matching search query",
                GlobalArgType.FunctionCategory.name: "external",
                GlobalArgType.ActionType.name: "list",
                GlobalArgType.ObjectSelector.name: "service",
            },
        ])

    def is_test_data_allowed(
        self,
        test_data_id: str,
    ) -> bool:
        if test_data_id in self.config_dict[test_data_ids_to_load_]:
            return True
        return False

    @staticmethod
    def populate_common_AccessType(data_envelopes: list):

        data_envelopes.extend([

            ############################################################################################################
            # `AccessType`: FS_24_50_40_64

            {
                envelope_payload_: {
                },
                ReservedArgType.EnvelopeClass.name: ServiceArgType.AccessType.name,
                ServiceArgType.AccessType.name: "ro",
            },
            {
                envelope_payload_: {
                },
                ReservedArgType.EnvelopeClass.name: ServiceArgType.AccessType.name,
                ServiceArgType.AccessType.name: "rw",
            },
        ])

    def populate_TD_63_37_05_36_default(
        self,
        data_envelopes: list,
    ):
        """
        Populates TD_63_37_05_36 # demo
        """
        if not self.is_test_data_allowed("TD_63_37_05_36"):
            return

        data_envelopes.extend([

            ############################################################################################################
            # TD_63_37_05_36 # demo: clusters

            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassCluster.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "dev-apac-upstream",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassCluster.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "dev-apac-downstream",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassCluster.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "emea",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "dev-emea-upstream",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassCluster.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "emea",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "dev-emea-downstream",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassCluster.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "amer",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "dev-amer-upstream",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassCluster.name,
                ServiceArgType.CodeMaturity.name: "qa",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "qa-apac-upstream",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassCluster.name,
                ServiceArgType.CodeMaturity.name: "qa",
                ServiceArgType.GeoRegion.name: "emea",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "qa-emea-downstream",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassCluster.name,
                ServiceArgType.CodeMaturity.name: "qa",
                ServiceArgType.GeoRegion.name: "amer",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "qa-amer-upstream",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassCluster.name,
                ServiceArgType.CodeMaturity.name: "qa",
                ServiceArgType.GeoRegion.name: "amer",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "qa-amer-downstream",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassCluster.name,
                ServiceArgType.CodeMaturity.name: "prod",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "prod-apac-upstream",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassCluster.name,
                ServiceArgType.CodeMaturity.name: "prod",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "prod-apac-downstream",
            },

            ############################################################################################################
            # TD_63_37_05_36 # demo: hosts

            # dev

            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassHost.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "dev-apac-upstream",
                ServiceArgType.DataCenter.name: "dc.01",
                ServiceArgType.HostName.name: "zxcv-du",
                ServiceArgType.IpAddress.name: "ip.192.168.1.1",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassHost.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "dev-apac-downstream",
                ServiceArgType.DataCenter.name: "dc.11",
                ServiceArgType.HostName.name: "zxcv-dd",
                ServiceArgType.IpAddress.name: "ip.172.16.1.2",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassHost.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "dev-apac-downstream",
                ServiceArgType.DataCenter.name: "dc.01",
                ServiceArgType.HostName.name: "poiu-dd",
                ServiceArgType.IpAddress.name: "ip.192.168.1.3",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassHost.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "emea",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "dev-emea-upstream",
                ServiceArgType.DataCenter.name: "dc.22",
                ServiceArgType.HostName.name: "asdf-du",
                ServiceArgType.IpAddress.name: "ip.172.16.2.1",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassHost.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "emea",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "dev-emea-downstream",
                ServiceArgType.DataCenter.name: "dc.02",
                ServiceArgType.HostName.name: "xcvb-dd",
                ServiceArgType.IpAddress.name: "ip.192.168.2.2",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassHost.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "amer",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "dev-amer-upstream",
                ServiceArgType.DataCenter.name: "dc.03",
                ServiceArgType.HostName.name: "qwer-du",
                ServiceArgType.IpAddress.name: "ip.192.168.3.1",
            },

            # qa

            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassHost.name,
                ServiceArgType.CodeMaturity.name: "qa",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "qa-apac-upstream",
                ServiceArgType.DataCenter.name: "dc.04",
                ServiceArgType.HostName.name: "hjkl-qu",
                ServiceArgType.IpAddress.name: "ip.192.168.4.1",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassHost.name,
                ServiceArgType.CodeMaturity.name: "qa",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "qa-apac-upstream",
                ServiceArgType.DataCenter.name: "dc.44",
                ServiceArgType.HostName.name: "poiu-qu",
                ServiceArgType.IpAddress.name: "ip.172.16.4.2",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassHost.name,
                ServiceArgType.CodeMaturity.name: "qa",
                ServiceArgType.GeoRegion.name: "amer",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "qa-amer-upstream",
                ServiceArgType.DataCenter.name: "dc.06",
                ServiceArgType.HostName.name: "rtyu-qu",
                ServiceArgType.IpAddress.name: "ip.192.168.6.1",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassHost.name,
                ServiceArgType.CodeMaturity.name: "qa",
                ServiceArgType.GeoRegion.name: "amer",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "qa-amer-upstream",
                ServiceArgType.DataCenter.name: "dc.06",
                ServiceArgType.HostName.name: "rt-qu",
                ServiceArgType.IpAddress.name: "ip.192.168.6.2",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassHost.name,
                ServiceArgType.CodeMaturity.name: "qa",
                ServiceArgType.GeoRegion.name: "amer",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "qa-amer-downstream",
                ServiceArgType.DataCenter.name: "dc.06",
                ServiceArgType.HostName.name: "sdfgh-qd",
                ServiceArgType.IpAddress.name: "ip.192.168.6.3",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassHost.name,
                ServiceArgType.CodeMaturity.name: "qa",
                ServiceArgType.GeoRegion.name: "amer",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "qa-amer-downstream",
                ServiceArgType.DataCenter.name: "dc.06",
                ServiceArgType.HostName.name: "sdfgb-qd",
                ServiceArgType.IpAddress.name: "ip.192.168.6.4",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassHost.name,
                ServiceArgType.CodeMaturity.name: "qa",
                ServiceArgType.GeoRegion.name: "amer",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "qa-amer-downstream",
                ServiceArgType.DataCenter.name: "dc.66",
                ServiceArgType.HostName.name: "sdfg-qd",
                ServiceArgType.IpAddress.name: "ip.172.16.6.5",
            },

            # prod

            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassHost.name,
                ServiceArgType.CodeMaturity.name: "prod",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "prod-apac-upstream",
                ServiceArgType.DataCenter.name: "dc.07",
                ServiceArgType.HostName.name: "qwer-pd-1",
                ServiceArgType.IpAddress.name: "ip.192.168.7.1",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassHost.name,
                ServiceArgType.CodeMaturity.name: "prod",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "prod-apac-upstream",
                ServiceArgType.DataCenter.name: "dc.07",
                ServiceArgType.HostName.name: "qwer-pd-3",
                ServiceArgType.IpAddress.name: "ip.192.168.7.2",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassHost.name,
                ServiceArgType.CodeMaturity.name: "prod",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "prod-apac-upstream",
                ServiceArgType.DataCenter.name: "dc.77",
                ServiceArgType.HostName.name: "qwer-pd-2",
                ServiceArgType.IpAddress.name: "ip.172.16.7.2",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassHost.name,
                ServiceArgType.CodeMaturity.name: "prod",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "prod-apac-downstream",
                ServiceArgType.DataCenter.name: "dc.07",
                ServiceArgType.HostName.name: "wert-pd-1",
                ServiceArgType.IpAddress.name: "ip.192.168.7.3",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassHost.name,
                ServiceArgType.CodeMaturity.name: "prod",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "prod-apac-downstream",
                ServiceArgType.DataCenter.name: "dc.07",
                ServiceArgType.HostName.name: "wert-pd-2",
                ServiceArgType.IpAddress.name: "ip.192.168.7.4",
            },

            ############################################################################################################
            # TD_63_37_05_36 # demo: services

            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassService.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "dev-apac-upstream",
                ServiceArgType.DataCenter.name: "dc.01",
                ServiceArgType.HostName.name: "zxcv-du",
                ServiceArgType.IpAddress.name: "ip.192.168.1.1",
                ServiceArgType.ServiceName.name: "s_a",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassService.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "dev-apac-upstream",
                ServiceArgType.DataCenter.name: "dc.01",
                ServiceArgType.HostName.name: "zxcv-du",
                ServiceArgType.IpAddress.name: "ip.192.168.1.1",
                ServiceArgType.ServiceName.name: "s_b",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassService.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "dev-apac-upstream",
                ServiceArgType.DataCenter.name: "dc.01",
                ServiceArgType.HostName.name: "zxcv-du",
                ServiceArgType.IpAddress.name: "ip.192.168.1.1",
                ServiceArgType.ServiceName.name: "s_c",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassService.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "dev-apac-downstream",
                ServiceArgType.DataCenter.name: "dc.11",
                ServiceArgType.HostName.name: "zxcv-dd",
                ServiceArgType.IpAddress.name: "ip.172.16.1.2",
                ServiceArgType.ServiceName.name: "tt",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassService.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "dev-apac-downstream",
                ServiceArgType.DataCenter.name: "dc.01",
                ServiceArgType.HostName.name: "poiu-dd",
                ServiceArgType.IpAddress.name: "ip.192.168.1.3",
                ServiceArgType.ServiceName.name: "xx",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassService.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "emea",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "dev-emea-upstream",
                ServiceArgType.DataCenter.name: "dc.22",
                ServiceArgType.HostName.name: "asdf-du",
                ServiceArgType.IpAddress.name: "ip.172.16.2.1",
                ServiceArgType.ServiceName.name: "s_a",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassService.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "emea",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "dev-emea-upstream",
                ServiceArgType.DataCenter.name: "dc.22",
                ServiceArgType.HostName.name: "asdf-du",
                ServiceArgType.IpAddress.name: "ip.172.16.2.1",
                ServiceArgType.ServiceName.name: "s_b",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassService.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "emea",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "dev-emea-downstream",
                ServiceArgType.DataCenter.name: "dc.02",
                ServiceArgType.HostName.name: "xcvb-dd",
                ServiceArgType.IpAddress.name: "ip.192.168.2.2",
                ServiceArgType.ServiceName.name: "xx",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassService.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "emea",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "dev-emea-downstream",
                ServiceArgType.DataCenter.name: "dc.02",
                ServiceArgType.HostName.name: "xcvb-dd",
                ServiceArgType.IpAddress.name: "ip.192.168.2.2",
                ServiceArgType.ServiceName.name: "zz",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassService.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "amer",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "dev-amer-upstream",
                ServiceArgType.DataCenter.name: "dc.03",
                ServiceArgType.HostName.name: "qwer-du",
                ServiceArgType.IpAddress.name: "ip.192.168.3.1",
                ServiceArgType.ServiceName.name: "s_a",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassService.name,
                ServiceArgType.CodeMaturity.name: "qa",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "qa-apac-upstream",
                ServiceArgType.DataCenter.name: "dc.04",
                ServiceArgType.HostName.name: "hjkl-qu",
                ServiceArgType.IpAddress.name: "ip.192.168.4.1",
                ServiceArgType.ServiceName.name: "s_a",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassService.name,
                ServiceArgType.CodeMaturity.name: "qa",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "qa-apac-upstream",
                ServiceArgType.DataCenter.name: "dc.04",
                ServiceArgType.HostName.name: "hjkl-qu",
                ServiceArgType.IpAddress.name: "ip.192.168.4.1",
                ServiceArgType.ServiceName.name: "s_b",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassService.name,
                ServiceArgType.CodeMaturity.name: "qa",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "qa-apac-upstream",
                ServiceArgType.DataCenter.name: "dc.44",
                ServiceArgType.HostName.name: "poiu-qu",
                ServiceArgType.IpAddress.name: "ip.172.16.4.2",
                ServiceArgType.ServiceName.name: "s_c",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassService.name,
                ServiceArgType.CodeMaturity.name: "qa",
                ServiceArgType.GeoRegion.name: "amer",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "qa-amer-upstream",
                ServiceArgType.DataCenter.name: "dc.06",
                ServiceArgType.HostName.name: "rtyu-qu",
                ServiceArgType.IpAddress.name: "ip.192.168.6.1",
                ServiceArgType.ServiceName.name: "s_a",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassService.name,
                ServiceArgType.CodeMaturity.name: "qa",
                ServiceArgType.GeoRegion.name: "amer",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "qa-amer-downstream",
                ServiceArgType.DataCenter.name: "dc.06",
                ServiceArgType.HostName.name: "sdfgh-qd",
                ServiceArgType.IpAddress.name: "ip.192.168.6.3",
                ServiceArgType.ServiceName.name: "tt1",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassService.name,
                ServiceArgType.CodeMaturity.name: "qa",
                ServiceArgType.GeoRegion.name: "amer",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "qa-amer-downstream",
                ServiceArgType.DataCenter.name: "dc.06",
                ServiceArgType.HostName.name: "sdfgb-qd",
                ServiceArgType.IpAddress.name: "ip.192.168.6.4",
                ServiceArgType.ServiceName.name: "xx",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassService.name,
                ServiceArgType.CodeMaturity.name: "prod",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "prod-apac-upstream",
                ServiceArgType.DataCenter.name: "dc.07",
                ServiceArgType.HostName.name: "qwer-pd-1",
                ServiceArgType.IpAddress.name: "ip.192.168.7.1",
                ServiceArgType.ServiceName.name: "s_a",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassService.name,
                ServiceArgType.CodeMaturity.name: "prod",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "prod-apac-upstream",
                ServiceArgType.DataCenter.name: "dc.07",
                ServiceArgType.HostName.name: "qwer-pd-1",
                ServiceArgType.IpAddress.name: "ip.192.168.7.1",
                ServiceArgType.ServiceName.name: "s_b",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassService.name,
                ServiceArgType.CodeMaturity.name: "prod",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "prod-apac-upstream",
                ServiceArgType.DataCenter.name: "dc.07",
                ServiceArgType.HostName.name: "qwer-pd-3",
                ServiceArgType.IpAddress.name: "ip.192.168.7.2",
                ServiceArgType.ServiceName.name: "s_c",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassService.name,
                ServiceArgType.CodeMaturity.name: "prod",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "prod-apac-upstream",
                ServiceArgType.DataCenter.name: "dc.77",
                ServiceArgType.HostName.name: "qwer-pd-2",
                ServiceArgType.IpAddress.name: "ip.172.16.7.2",
                ServiceArgType.ServiceName.name: "s_a",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassService.name,
                ServiceArgType.CodeMaturity.name: "prod",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "prod-apac-upstream",
                ServiceArgType.DataCenter.name: "dc.77",
                ServiceArgType.HostName.name: "qwer-pd-2",
                ServiceArgType.IpAddress.name: "ip.172.16.7.2",
                ServiceArgType.ServiceName.name: "s_b",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassService.name,
                ServiceArgType.CodeMaturity.name: "prod",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "upstream",
                ServiceArgType.ClusterName.name: "prod-apac-upstream",
                ServiceArgType.DataCenter.name: "dc.77",
                ServiceArgType.HostName.name: "qwer-pd-2",
                ServiceArgType.IpAddress.name: "ip.172.16.7.2",
                ServiceArgType.ServiceName.name: "s_c",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassService.name,
                ServiceArgType.CodeMaturity.name: "prod",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "prod-apac-downstream",
                ServiceArgType.DataCenter.name: "dc.07",
                ServiceArgType.HostName.name: "wert-pd-1",
                ServiceArgType.IpAddress.name: "ip.192.168.7.3",
                ServiceArgType.ServiceName.name: "tt1",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassService.name,
                ServiceArgType.CodeMaturity.name: "prod",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "prod-apac-downstream",
                ServiceArgType.DataCenter.name: "dc.07",
                ServiceArgType.HostName.name: "wert-pd-2",
                ServiceArgType.IpAddress.name: "ip.192.168.7.4",
                ServiceArgType.ServiceName.name: "tt2",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassService.name,
                ServiceArgType.CodeMaturity.name: "prod",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "prod-apac-downstream",
                ServiceArgType.DataCenter.name: "dc.07",
                ServiceArgType.HostName.name: "wert-pd-2",
                ServiceArgType.IpAddress.name: "ip.192.168.7.4",
                ServiceArgType.ServiceName.name: "xx",
            },
        ])

    def populate_TD_76_09_29_31_overlapped(
        self,
        data_envelopes: list,
    ):
        if not self.is_test_data_allowed("TD_76_09_29_31"):
            return

        data_envelopes.extend([

            ############################################################################################################
            # TD_76_09_29_31 # overlapped: clusters

            {
                envelope_payload_: {
                },
                test_data_: "TD_76_09_29_31",  # overlapped
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassCluster.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "amer",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "dev-amer-downstream",
            },

            {
                envelope_payload_: {
                },
                test_data_: "TD_76_09_29_31",  # overlapped
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassCluster.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "emea",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "dev-emea-downstream",
            },

            ############################################################################################################
            # TD_76_09_29_31 # overlapped: hosts

            {
                envelope_payload_: {
                },
                test_data_: "TD_76_09_29_31",  # overlapped
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassHost.name,
                ServiceArgType.CodeMaturity.name: "dev",
                # TD_76_09_29_31 GeoRegion set overlaps with HostName set:
                ServiceArgType.GeoRegion.name: "amer",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "dev-amer-downstream",
                # TD_76_09_29_31 HostName set overlaps with GeoRegion set:
                ServiceArgType.HostName.name: "amer",
            },

            {
                envelope_payload_: {
                },
                test_data_: "TD_76_09_29_31",  # overlapped
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassHost.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "amer",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "dev-amer-downstream",
                # HostName is intentionally mathing another HostName from different GeoRegion (also named alike):
                ServiceArgType.HostName.name: "emea",
            },

            {
                envelope_payload_: {
                },
                test_data_: "TD_76_09_29_31",  # overlapped
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassHost.name,
                ServiceArgType.CodeMaturity.name: "dev",
                # TD_76_09_29_31 GeoRegion set overlaps with HostName set:
                ServiceArgType.GeoRegion.name: "emea",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "dev-emea-downstream",
                # TD_76_09_29_31 HostName set overlaps with GeoRegion set:
                ServiceArgType.HostName.name: "emea",
            },

            {
                envelope_payload_: {
                },
                test_data_: "TD_76_09_29_31",  # overlapped
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassHost.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "emea",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "dev-emea-downstream",
                # HostName is intentionally mathing another HostName from different GeoRegion (also named alike):
                ServiceArgType.HostName.name: "amer",
            },
        ])

    def populate_TD_38_03_48_51_large_generated(
        self,
        data_envelopes: list,
    ):
        """
        TD_38_03_48_51: generate large data set
        """
        if not self.is_test_data_allowed("TD_38_03_48_51"):
            return

        for code_maturity in ["cm" + str(cmn) for cmn in range(0, 10)]:
            for geo_region in ["gr" + str(grn) for grn in range(0, 10)]:
                for flow_stage in ["fs" + str(fsn) for fsn in range(0, 10)]:

                    cluster_name = f"{code_maturity}-{geo_region}-{flow_stage}"

                    eprint(f"loading cluster_name={cluster_name}...")

                    #####################################################################################################
                    # clusters

                    generated_cluster = {
                        envelope_payload_: {
                        },
                        test_data_: "TD_38_03_48_51",  # large generated
                        ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassCluster.name,
                        ServiceArgType.CodeMaturity.name: code_maturity,
                        ServiceArgType.GeoRegion.name: geo_region,
                        ServiceArgType.FlowStage.name: flow_stage,
                        ServiceArgType.ClusterName.name: cluster_name,
                    }

                    data_envelopes.append(generated_cluster)

                    for host_name in ["hs" + str(hsn) for hsn in range(0, 10)]:

                        ################################################################################################
                        # hosts

                        generated_host = {
                            envelope_payload_: {
                            },
                            test_data_: "TD_38_03_48_51",  # large generated
                            ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassHost.name,
                            ServiceArgType.CodeMaturity.name: code_maturity,
                            ServiceArgType.GeoRegion.name: geo_region,
                            ServiceArgType.FlowStage.name: flow_stage,
                            ServiceArgType.ClusterName.name: cluster_name,
                            ServiceArgType.HostName.name: host_name,
                        }

                        data_envelopes.append(generated_host)

                        for service_name in ["sn{:02d}".format(snn) for snn in range(0, 10)]:
                            ############################################################################################################
                            # services

                            generated_service = {
                                envelope_payload_: {
                                },
                                test_data_: "TD_38_03_48_51",  # large generated
                                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassService.name,
                                ServiceArgType.CodeMaturity.name: code_maturity,
                                ServiceArgType.GeoRegion.name: geo_region,
                                ServiceArgType.FlowStage.name: flow_stage,
                                ServiceArgType.ClusterName.name: cluster_name,
                                ServiceArgType.HostName.name: host_name,
                                ServiceArgType.ServiceName.name: service_name,
                            }

                            data_envelopes.append(generated_service)

    def populate_TD_43_24_76_58_single(
        self,
        data_envelopes: list,
    ):
        if not self.is_test_data_allowed("TD_43_24_76_58"):
            return

        data_envelopes.extend([

            ############################################################################################################
            # TD_43_24_76_58 # single: clusters

            {
                envelope_payload_: {
                },
                test_data_: "TD_43_24_76_58",  # single
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassCluster.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "dev-apac-downstream",
            },

            {
                envelope_payload_: {
                },
                test_data_: "TD_43_24_76_58",  # single
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassCluster.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "emea",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "dev-emea-downstream",
            },

            ############################################################################################################
            # TD_43_24_76_58 # single: hosts

            {
                envelope_payload_: {
                },
                test_data_: "TD_43_24_76_58",  # single
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassHost.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "dev-apac-downstream",
                ServiceArgType.HostName.name: "qwer",
            },

            {
                envelope_payload_: {
                },
                test_data_: "TD_43_24_76_58",  # single
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassHost.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "apac",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "dev-apac-downstream",
                ServiceArgType.HostName.name: "asdf",
            },

            {
                envelope_payload_: {
                },
                test_data_: "TD_43_24_76_58",  # single
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassHost.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "emea",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "dev-emea-downstream",
                ServiceArgType.HostName.name: "qwer",
            },

            {
                envelope_payload_: {
                },
                test_data_: "TD_43_24_76_58",  # single
                ReservedArgType.EnvelopeClass.name: ServiceEnvelopeClass.ClassHost.name,
                ServiceArgType.CodeMaturity.name: "dev",
                ServiceArgType.GeoRegion.name: "emea",
                ServiceArgType.FlowStage.name: "downstream",
                ServiceArgType.ClusterName.name: "dev-emea-downstream",
                ServiceArgType.HostName.name: "asdf",
            },
        ])



from argrelay.plugin_interp.FirstArgInterp import FirstArgInterp
from argrelay.plugin_interp.FirstArgInterpFactoryConfigSchema import (
    first_arg_interp_factory_config_desc,
    first_arg_vals_to_next_interp_factory_ids_,
)
from argrelay.plugin_interp.TreePathInterpFactory import TreePathInterpFactory
from argrelay.plugin_interp.TreePathInterpFactoryConfigSchema import (
    tree_path_interp_factory_config_desc,
    interp_selector_tree_,
)
from argrelay.runtime_context.InterpContext import InterpContext


class FirstArgInterpFactory(TreePathInterpFactory):
    """
    `FirstArgInterp` was re-implemented in terms of `TreePathInterp` (FS_01_89_09_24).

    Implements FS_42_76_93_51.
    """

    def __init__(
        self,
        plugin_instance_id: str,
        config_dict: dict,
    ):
        first_arg_interp_factory_config_desc.validate_dict(config_dict)
        converted_config = convert_FirstArgInterpConfig_to_TreePathInterpFactoryConfig(config_dict)
        tree_path_interp_factory_config_desc.validate_dict(converted_config)
        super().__init__(
            plugin_instance_id,
            converted_config,
        )

    def create_interp(
        self,
        interp_ctx: InterpContext,
    ) -> FirstArgInterp:
        return FirstArgInterp(
            self.plugin_instance_id,
            self.config_dict,
            interp_ctx,
        )


def convert_FirstArgInterpConfig_to_TreePathInterpFactoryConfig(config_dict: dict) -> dict:
    """
    `FirstArgInterp` was re-implemented in terms of `TreePathInterp` (FS_01_89_09_24).
    """
    interp_selector_tree = {}
    for command_id, plugin_instance_id in config_dict[first_arg_vals_to_next_interp_factory_ids_].items():
        interp_selector_tree[command_id] = plugin_instance_id

    output_dict = {
        interp_selector_tree_: interp_selector_tree
    }

    tree_path_interp_factory_config_desc.validate_dict(output_dict)

    return output_dict



from argrelay.client_command_local.AbstractLocalClientCommand import AbstractLocalClientCommand
from argrelay.misc_helper.ElapsedTime import ElapsedTime
from argrelay.relay_client.AbstractClientCommandFactory import AbstractClientCommandFactory
from argrelay.relay_server.LocalServer import LocalServer
from argrelay.runtime_data.ServerConfig import ServerConfig
from argrelay.schema_config_core_server.ServerConfigSchema import server_config_desc
from argrelay.server_spec.const_int import (
    DESCRIBE_LINE_ARGS_PATH,
    RELAY_LINE_ARGS_PATH,
    PROPOSE_ARG_VALUES_PATH,
)


class LocalClientCommandFactory(AbstractClientCommandFactory):

    def __init__(self):
        self._start_local_server()

    def _start_local_server(self):
        self.server_config: ServerConfig = server_config_desc.from_default_file()
        ElapsedTime.measure("after_server_config_load")
        self.local_server: LocalServer = LocalServer(self.server_config)
        self.local_server.start_local_server()
        ElapsedTime.measure("after_local_server_start")

    def create_command_by_server_path(self, server_path: str) -> AbstractLocalClientCommand:
        if server_path == DESCRIBE_LINE_ARGS_PATH:
            from argrelay.client_command_local.DescribeLineArgsLocalClientCommand import (
                DescribeLineArgsLocalClientCommand
            )
            return DescribeLineArgsLocalClientCommand(
                self.server_config,
                self.local_server,
            )
        if server_path == PROPOSE_ARG_VALUES_PATH:
            from argrelay.client_command_local.ProposeArgValuesLocalClientCommand import (
                ProposeArgValuesLocalClientCommand
            )
            return ProposeArgValuesLocalClientCommand(
                self.server_config,
                self.local_server,
            )
        if server_path == RELAY_LINE_ARGS_PATH:
            from argrelay.client_command_local.RelayLineArgsLocalClientCommand import (
                RelayLineArgsLocalClientCommand
            )
            return RelayLineArgsLocalClientCommand(
                self.server_config,
                self.local_server,
            )
        raise RuntimeError



from argrelay.relay_client.AbstractClientCommandFactory import AbstractClientCommandFactory
from argrelay.runtime_data.ClientConfig import ClientConfig
from argrelay.server_spec.const_int import (
    DESCRIBE_LINE_ARGS_PATH,
    RELAY_LINE_ARGS_PATH,
    PROPOSE_ARG_VALUES_PATH,
)


class RemoteClientCommandFactory(AbstractClientCommandFactory):

    def __init__(
        self,
        client_config: ClientConfig,
    ):
        self.client_config: ClientConfig = client_config

    # noinspection PyMethodMayBeStatic
    def create_command_by_server_path(self, server_path: str) -> "AbstractRemoteClientCommand":
        if server_path == DESCRIBE_LINE_ARGS_PATH:
            from argrelay.client_command_remote.DescribeLineArgsRemoteClientCommand import (
                DescribeLineArgsRemoteClientCommand,
            )
            return DescribeLineArgsRemoteClientCommand(self.client_config.connection_config)
        if server_path == PROPOSE_ARG_VALUES_PATH:
            from argrelay.client_command_remote.ProposeArgValuesRemoteClientCommand import (
                ProposeArgValuesRemoteClientCommand,
            )
            return ProposeArgValuesRemoteClientCommand(self.client_config.connection_config)
        if server_path == RELAY_LINE_ARGS_PATH:
            from argrelay.client_command_remote.RelayLineArgsRemoteClientCommand import (
                RelayLineArgsRemoteClientCommand,
            )
            return RelayLineArgsRemoteClientCommand(self.client_config.connection_config)
        raise RuntimeError



from __future__ import annotations

import os
from dataclasses import dataclass, field

from argrelay.enum_desc.CompType import CompType
from argrelay.enum_desc.RunMode import RunMode
from argrelay.enum_desc.TermColor import TermColor
from argrelay.misc_helper import eprint
from argrelay.runtime_context.RequestContext import RequestContext


@dataclass(frozen = True)
class InputContext(RequestContext):
    """
    Immutable input data
    """

    comp_key: str = field()
    run_mode: RunMode = field()

    @classmethod
    def from_request_context(
        cls,
        request_ctx: RequestContext,
        run_mode: RunMode,
        comp_key: str,
    ):
        return cls(
            command_line = request_ctx.command_line,
            cursor_cpos = request_ctx.cursor_cpos,
            comp_type = request_ctx.comp_type,
            is_debug_enabled = request_ctx.is_debug_enabled,
            run_mode = run_mode,
            comp_key = comp_key,
        )

    @classmethod
    def from_env(cls, argv: list[str]):
        """
        See Bash docs on these env var:
        https://www.gnu.org/software/bash/manual/html_node/Bash-Variables.html
        """
        if "COMP_LINE" in os.environ:
            # If "COMP_LINE" env var exists, the call is for CompletionMode:
            command_line = os.environ["COMP_LINE"]
            cursor_cpos = int(os.environ["COMP_POINT"])
            comp_type = CompType(int(os.environ["COMP_TYPE"]))
            comp_key = os.environ["COMP_KEY"]
            if comp_type == CompType.DescribeArgs:
                # FS_23_62_89_43:
                # To process tangent token in case of `CompType.DescribeArgs`, use `RunMode.InvocationMode`:
                run_mode = RunMode.InvocationMode
            else:
                run_mode = RunMode.CompletionMode
        else:
            # If no "COMP_LINE" env var, the call is for InvocationMode:
            run_mode = RunMode.InvocationMode
            argv = [os.path.basename(argv[0])] + argv[1:]
            command_line = " ".join(argv)
            cursor_cpos = len(command_line)
            comp_type = CompType.InvokeAction
            comp_key = str(0)

        is_debug_enabled = "ARGRELAY_DEBUG" in os.environ

        return cls(
            command_line = command_line,
            cursor_cpos = cursor_cpos,
            comp_type = comp_type,
            is_debug_enabled = is_debug_enabled,
            comp_key = comp_key,
            run_mode = run_mode,
        )

    def print_debug(self, end_str: str = "\n") -> None:
        if not self.is_debug_enabled:
            return
        eprint(TermColor.DEBUG.value, end = "")
        eprint(f"\"{self.command_line}\"", end = " ")
        eprint(f"cursor_cpos: {self.cursor_cpos}", end = " ")
        eprint(f"run_mode: {self.run_mode}", end = " ")
        eprint(TermColor.RESET.value, end = end_str)



from marshmallow import fields


class FilteredDict(fields.Dict):
    """
    Dict field which does not serialize specific fields:
    https://github.com/marshmallow-code/marshmallow/issues/494#issuecomment-390495145
    """

    def __init__(self, filtered_keys, **kwargs):
        super(FilteredDict, self).__init__(**kwargs)
        self.filtered_keys = filtered_keys

    def _serialize(self, value, attr, obj, **kwargs):
        value = super(FilteredDict, self)._serialize(value, attr, obj)
        if value:
            return {key: value[key] for key in value if key not in self.filtered_keys}
        else:
            return value

    def _deserialize(self, value, attr, data, **kwargs):
        value = super(FilteredDict, self)._deserialize(value, attr, data)
        return {key: value[key] for key in value if key not in self.filtered_keys}



from argrelay.client_command_remote.AbstractRemoteClientCommand import AbstractRemoteClientCommand
from argrelay.handler_response.RelayLineArgsClientResponseHandler import RelayLineArgsClientResponseHandler
from argrelay.runtime_data.ConnectionConfig import ConnectionConfig
from argrelay.schema_response.InvocationInputSchema import invocation_input_desc
from argrelay.server_spec.const_int import RELAY_LINE_ARGS_PATH


class RelayLineArgsRemoteClientCommand(AbstractRemoteClientCommand):

    def __init__(
        self,
        connection_config: ConnectionConfig,
    ):
        super().__init__(
            server_path = RELAY_LINE_ARGS_PATH,
            connection_config = connection_config,
            response_handler = RelayLineArgsClientResponseHandler(),
            response_schema = invocation_input_desc.dict_schema,
        )



from __future__ import annotations

from argrelay.enum_desc.InterpStep import InterpStep
from argrelay.plugin_interp.AbstractInterp import AbstractInterp
from argrelay.plugin_interp.NoopInterpFactory import NoopInterpFactory
from argrelay.plugin_interp.TreePathInterpFactoryConfigSchema import interp_selector_tree_
from argrelay.runtime_context.InterpContext import InterpContext

default_tree_leaf_ = ""


def fetch_tree_node(
    tree_dict: dict,
    node_path: list[str],
):
    """
    Fetches tree node by node path.
    """

    curr_node_value = tree_dict
    for node_path_part in node_path:
        if isinstance(curr_node_value, dict) and node_path_part in curr_node_value:
            curr_node_value = curr_node_value[node_path_part]
        else:
            return None
    return curr_node_value


class TreePathInterp(AbstractInterp):
    """
    Implements FS_01_89_09_24.
    """

    def __init__(
        self,
        interp_factory_id: str,
        config_dict: dict,
        interp_ctx: InterpContext,
    ):
        super().__init__(
            interp_factory_id,
            config_dict,
            interp_ctx,
        )
        self.interp_selector_tree: dict = config_dict[interp_selector_tree_]
        self.interp_factory_id: str = interp_factory_id
        self.node_path = []

        # Token with ipos = 0 is the command name eaten by `FirstArgInterp`:
        self.base_token_ipos: int = 1

    def consume_pos_args(self) -> None:
        """
        Consumes heading args in `unconsumed_tokens` according to the `interp_selector_tree`.
        """

        self.node_path = []
        curr_sub_tree = self.interp_selector_tree
        while True:
            if isinstance(curr_sub_tree, str):
                # Tree leaf is reached - use it:
                self.interp_factory_id = curr_sub_tree
                return

            if not self.interp_ctx.unconsumed_tokens:
                self.set_default_factory_id(curr_sub_tree)
                return

            # Always consume next unconsumed token:
            # TODO: Is this assumption valid/safe that next unconsumed `ipos` is in the order it appears on command line?
            curr_token_ipos = self.interp_ctx.unconsumed_tokens[0]
            curr_token_value = self.interp_ctx.parsed_ctx.all_tokens[curr_token_ipos]
            assert self.is_pos_arg(curr_token_ipos)

            if isinstance(curr_sub_tree, dict):
                if curr_token_value in curr_sub_tree:
                    # Consume one more arg into path:
                    self.node_path.append(curr_token_value)
                    self.interp_ctx.consumed_tokens.append(curr_token_ipos)
                    del self.interp_ctx.unconsumed_tokens[0]
                    curr_sub_tree = curr_sub_tree[curr_token_value]
                    continue
                else:
                    self.set_default_factory_id(curr_sub_tree)
                    return
            else:
                raise LookupError()

    def set_default_factory_id(self, curr_sub_tree):
        # Impossible to consume more arg - use default of the current sub-tree:
        if default_tree_leaf_ in curr_sub_tree:
            self.interp_factory_id = curr_sub_tree[default_tree_leaf_]
        else:
            self.interp_factory_id = NoopInterpFactory.__name__

    def try_iterate(self) -> InterpStep:
        return InterpStep.NextInterp

    def next_interp(self) -> "AbstractInterp":
        return self.interp_ctx.create_next_interp(self.interp_factory_id)

    def propose_arg_completion(self) -> None:
        """
        Suggest based on current sub-tree from config.
        """

        if not self.is_eligible_for_suggestion():
            return

        curr_sub_tree = fetch_tree_node(
            self.interp_selector_tree,
            self.node_path,
        )
        if isinstance(curr_sub_tree, dict):
            proposed_values = [
                x for x in curr_sub_tree.keys()
                if (
                    isinstance(x, str)
                    and
                    x != default_tree_leaf_
                    and
                    # FS_32_05_46_00: using `startwith`:
                    x.startswith(self.interp_ctx.parsed_ctx.tan_token_l_part)
                )
            ]
            self.interp_ctx.comp_suggestions.extend(proposed_values)

    def is_eligible_for_suggestion(self):
        """
        Suggesting anything is possible only if there is no other tokens after those recorded in `node_path`
        are available for consumption by subsequent interpreters.
        """

        if len(self.interp_ctx.unconsumed_tokens) == 0:
            return True
        else:
            return False

        # TODO: Clean up code below or take into account tangent token (ipos cursor position).
        #       Note that there is no ipos cursor position at the moment.
        #       It has to be computed with help of FS_23_62_89_43 tangent token
        #       (which might be a surrogate one if cursor does nto touch non-whitespace chars).
        remaining_consumed_tokens = self.interp_ctx.consumed_tokens.copy()

        # Remove everything until `base_token_ipos`:
        for token_ipos in range(0, self.base_token_ipos):
            assert token_ipos in remaining_consumed_tokens
            remaining_consumed_tokens.remove(token_ipos)

        # Remove everything eaten into `node_path`:
        token_ipos = self.base_token_ipos
        for node_path_token in self.node_path:
            assert token_ipos in remaining_consumed_tokens
            remaining_consumed_tokens.remove(token_ipos)
            token_ipos += 1

        # Is there anything else?
        return len(remaining_consumed_tokens) == 0



from argrelay.client_command_local.AbstractLocalClientCommand import AbstractLocalClientCommand
from argrelay.handler_request.RelayLineArgsServerRequestHandler import RelayLineArgsServerRequestHandler
from argrelay.handler_response.RelayLineArgsClientResponseHandler import RelayLineArgsClientResponseHandler
from argrelay.relay_server.LocalServer import LocalServer
from argrelay.runtime_data.ServerConfig import ServerConfig


class RelayLineArgsLocalClientCommand(AbstractLocalClientCommand):

    def __init__(
        self,
        server_config: ServerConfig,
        local_server: LocalServer,
    ):
        super().__init__(
            server_config = server_config,
            local_server = local_server,
            request_handler = RelayLineArgsServerRequestHandler(local_server),
            response_handler = RelayLineArgsClientResponseHandler(),
        )



from enum import Enum, auto


class RunMode(Enum):
    """
    `RunMode` originates on the client side because of different user actions:
    *   `CompletionMode`: user hits `Tab` trying to complete command line arg.
    *   `InvocationMode`: user hits `Enter` trying to execute the command line.
    Request to `argrelay` server indicates the two different modes:
    *   `CompletionMode` - when shell asks `argrelay` server for suggestions to complete one command line arg.
    *   `InvocationMode` - when already executed command uses `argrelay` to determine values for all args.
    In other words:
    *   `CompletionMode` = (proposing options) collects all possible value options for the single arg under cursor.
    *   `InvocationMode` = (exercising options) selects only one value for specific arg, but does it for every arg.
    Also:
    *   `CompletionMode` excludes tangent token ("touched" by the cursor) as input to provide completion for it.
    *   `InvocationMode` includes tangent token as input arg to execute selected function.
    """

    CompletionMode = auto()
    InvocationMode = auto()

    def __str__(self):
        return self.name



from marshmallow import Schema, fields, RAISE, post_load

from argrelay.misc_helper.TypeDesc import TypeDesc
from argrelay.relay_server.QueryCacheConfig import QueryCacheConfig

enable_query_cache_ = "enable_query_cache"
query_cache_ttl_sec_ = "query_cache_ttl_sec"
query_cache_max_size_bytes_ = "query_cache_max_size_bytes"


class QueryCacheConfigSchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    enable_query_cache = fields.Boolean()

    query_cache_ttl_sec = fields.Integer()

    query_cache_max_size_bytes = fields.Integer()

    @post_load
    def make_object(self, input_dict, **kwargs):
        return QueryCacheConfig(
            enable_query_cache = input_dict[enable_query_cache_],
            query_cache_ttl_sec = input_dict[query_cache_ttl_sec_],
            query_cache_max_size_bytes = input_dict[query_cache_max_size_bytes_],
        )


query_cache_config_desc = TypeDesc(
    dict_schema = QueryCacheConfigSchema(),
    ref_name = QueryCacheConfigSchema.__name__,
    dict_example = {
        enable_query_cache_: True,
        query_cache_ttl_sec_: 60,
        query_cache_max_size_bytes_: 1024 * 1024,
    },
    default_file_path = "",
)



from __future__ import annotations

from argrelay.enum_desc.SpecialFunc import SpecialFunc
from argrelay.plugin_delegator.AbstractDelegator import AbstractDelegator, get_data_envelopes
from argrelay.plugin_delegator.InvocationInput import InvocationInput
from argrelay.plugin_interp.AbstractInterp import AbstractInterp
from argrelay.relay_server.LocalServer import LocalServer
from argrelay.runtime_context.InterpContext import InterpContext, function_envelope_ipos_
from argrelay.runtime_context.SearchControl import SearchControl
from argrelay.schema_config_interp.DataEnvelopeSchema import envelope_id_, instance_data_
from argrelay.schema_config_interp.FunctionEnvelopeInstanceDataSchema import delegator_plugin_instance_id_

next_interp_plugin_instance_id_ = "next_interp_plugin_instance_id"


class InterceptDelegator(AbstractDelegator):

    def __init__(
        self,
        plugin_instance_id: str,
        config_dict: dict,
    ):
        super().__init__(
            plugin_instance_id,
            config_dict,
        )

    def run_search_control(
        self,
        function_data_envelope: dict,
    ) -> list[SearchControl]:
        # Nothing to search (only if next interpreter needs more, but this one is done):
        return []

    def run_interp_control(
        self,
        curr_interp: AbstractInterp,
    ) -> str:
        return self.config_dict[next_interp_plugin_instance_id_]

    def run_invoke_control(
        self,
        interp_ctx: InterpContext,
        local_server: LocalServer,
    ) -> InvocationInput:
        assert interp_ctx.is_funct_found(), "the (first) function envelope must be found"

        # TODO: Fail (send to ErrorDelegator) if next function is not specified -
        #       showing the payload in this case is misleading.
        function_envelope = interp_ctx.envelope_containers[function_envelope_ipos_]
        delegator_plugin_instance_id = function_envelope.data_envelope[instance_data_][delegator_plugin_instance_id_]
        invocation_input = InvocationInput(
            all_tokens = interp_ctx.parsed_ctx.all_tokens,
            consumed_tokens = interp_ctx.consumed_tokens,
            delegator_plugin_entry = local_server.server_config.plugin_dict[delegator_plugin_instance_id],
            data_envelopes = get_data_envelopes(interp_ctx),
            custom_plugin_data = {},
        )
        return invocation_input

    @staticmethod
    def invoke_action(invocation_input: InvocationInput):
        if invocation_input.data_envelopes[function_envelope_ipos_][envelope_id_] == SpecialFunc.intercept_func.name:
            # TODO: Print without first function `data_envelope` belonging to `intercept` function:
            print(invocation_input)



# Keep minimal import on start:
from argrelay.misc_helper import get_config_path
from argrelay.misc_helper.ElapsedTime import ElapsedTime

ElapsedTime.measure("after_program_entry")


def main():
    # Initial imports - see `completion_perf_notes.md`.
    import sys
    from argrelay.runtime_context.InputContext import InputContext
    ElapsedTime.measure("after_initial_imports")

    file_path = get_config_path("argrelay.client.json")
    client_config = load_client_config(file_path)
    ElapsedTime.measure("after_loading_client_config")

    input_ctx = InputContext.from_env(sys.argv)

    if client_config.use_local_requests:
        from argrelay.relay_client.LocalClient import LocalClient

        command_obj = make_request(LocalClient(client_config), input_ctx)
    else:
        from argrelay.relay_client.RemoteClient import RemoteClient

        command_obj = make_request(RemoteClient(client_config), input_ctx)

    ElapsedTime.measure("on_exit")
    if input_ctx.is_debug_enabled:
        ElapsedTime.print_all()

    return command_obj


def load_client_config(file_path):
    import json
    with open(file_path) as config_file:
        client_config_dict = json.load(config_file)
    client_config = client_config_dict_to_object(client_config_dict)
    return client_config


def client_config_dict_to_object(client_config_dict):
    from argrelay.runtime_data.ClientConfig import ClientConfig
    from argrelay.runtime_data.ConnectionConfig import ConnectionConfig
    client_config = ClientConfig(
        # Not importing constants like `use_local_requests_` from *Schema for performance:
        use_local_requests = client_config_dict["use_local_requests"],
        connection_config = ConnectionConfig(
            server_host_name = client_config_dict["connection_config"]["server_host_name"],
            server_port_number = client_config_dict["connection_config"]["server_port_number"],
        ),
    )
    return client_config


def make_request(abstract_client, input_ctx) -> "AbstractClientCommand":
    ElapsedTime.measure("before_client_invocation")
    return abstract_client.make_request(input_ctx)


if __name__ == "__main__":
    main()



from flask import Blueprint
from flask import render_template

from argrelay.relay_server.GuiBannerConfig import GuiBannerConfig
from argrelay.server_spec.const_int import (
    ARGRELAY_GUI_PATH,
    API_SPEC_PATH,
    API_DOCS_PATH,
)


def create_blueprint_gui(
    argrelay_version: str,
    gui_banner_config: GuiBannerConfig,
):
    blueprint_gui = Blueprint(
        name = "blueprint_gui",
        import_name = __name__,
    )

    # TODO: test and think if it is the right place
    @blueprint_gui.route(ARGRELAY_GUI_PATH)
    def basic_ui():
        return render_template(
            "argrelay_main.html",
            argrelay_version = argrelay_version,
            argrelay_api_docs_path = API_DOCS_PATH,
            argrelay_api_spec_path = API_SPEC_PATH,
            header_html = gui_banner_config.header_html,
            footer_html = gui_banner_config.footer_html,
        )

    return blueprint_gui



from marshmallow import Schema, fields, RAISE, post_load, pre_dump

from argrelay.enum_desc.CompType import CompType
from argrelay.misc_helper import ensure_value_is_enum
from argrelay.misc_helper.TypeDesc import TypeDesc
from argrelay.runtime_context.RequestContext import RequestContext

command_line_ = "command_line"
cursor_cpos_ = "cursor_cpos"
comp_type_ = "comp_type"
is_debug_enabled_ = "is_debug_enabled"

_sample_command_line = "some_command goto service "

_request_context_example = {
    command_line_: _sample_command_line,
    cursor_cpos_: len(_sample_command_line),
    comp_type_: CompType.PrefixShown.name,
    is_debug_enabled_: True,
}


class RequestContextSchema(Schema):
    class Meta:
        unknown = RAISE
        ordered = True

    command_line = fields.String(
        required = True,
        metadata = {
            "example": _request_context_example[command_line_],
        },
    )
    cursor_cpos = fields.Integer(
        required = True,
        metadata = {
            "description": "Cursor position within command line (0 = before the first char)",
            "example": _request_context_example[cursor_cpos_],
        },
    )
    comp_type = fields.Enum(
        CompType,
        required = True,
        metadata = {
            "description": (
                "Name for a completion type - see " + CompType.__name__ + " enum " +
                "(which maps into ASCII char sent by Bash to completion callback to indicate completion type)"
            ),
            "example": _request_context_example[comp_type_],
        },
    )
    is_debug_enabled = fields.Boolean(
        required = True,
        metadata = {
            "description": "Enable extra debug output",
            "example": _request_context_example[is_debug_enabled_],
        },
    )

    @pre_dump
    def make_dict(self, input_object: RequestContext, **kwargs):
        # TODO: figure out to populate all automatically and reduce duplication - this is error-prone:
        if isinstance(input_object, RequestContext):
            return {
                command_line_: input_object.command_line,
                cursor_cpos_: input_object.cursor_cpos,
                comp_type_: ensure_value_is_enum(input_object.comp_type, CompType),
                is_debug_enabled_: input_object.is_debug_enabled,
            }
        else:
            # Assuming it is as dict:
            return {
                command_line_: input_object[command_line_],
                cursor_cpos_: input_object[cursor_cpos_],
                comp_type_: ensure_value_is_enum(input_object[comp_type_], CompType),
                is_debug_enabled_: input_object[is_debug_enabled_],
            }

    @post_load
    def make_object(self, input_dict, **kwargs):
        return RequestContext(
            command_line = input_dict[command_line_],
            cursor_cpos = input_dict[cursor_cpos_],
            comp_type = ensure_value_is_enum(input_dict[comp_type_], CompType),
            is_debug_enabled = input_dict[is_debug_enabled_],
        )


request_context_desc = TypeDesc(
    dict_schema = RequestContextSchema(),
    ref_name = RequestContextSchema.__name__,
    dict_example = _request_context_example,
    default_file_path = "",
)



from __future__ import annotations

from marshmallow import Schema, RAISE, fields, post_load

from argrelay.misc_helper.TypeDesc import TypeDesc
from argrelay.runtime_data.StaticData import StaticData
from argrelay.schema_config_interp.DataEnvelopeSchema import data_envelope_desc, mongo_id_
from argrelay.schema_response.FilteredDict import FilteredDict

first_interp_factory_id_ = "first_interp_factory_id"
known_arg_types_ = "known_arg_types"
data_envelopes_ = "data_envelopes"


class StaticDataSchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    first_interp_factory_id = fields.String(
        required = True,
    )

    known_arg_types = fields.List(
        fields.String(),
        required = True,
    )

    data_envelopes = fields.List(
        FilteredDict(
            filtered_keys = [mongo_id_]
        ),
        required = True,
    )

    @post_load
    def make_object(self, input_dict, **kwargs):
        return StaticData(
            first_interp_factory_id = input_dict[first_interp_factory_id_],
            known_arg_types = input_dict[known_arg_types_],
            data_envelopes = input_dict[data_envelopes_],
        )


static_data_desc = TypeDesc(
    dict_schema = StaticDataSchema(),
    ref_name = StaticDataSchema.__name__,
    dict_example = {
        first_interp_factory_id_: "SomeInterp",
        known_arg_types_: [
            "SomeTypeA",
            "SomeTypeB",
        ],
        data_envelopes_: [
            data_envelope_desc.dict_example,
        ],
    },
    default_file_path = "",
)



from argrelay.plugin_interp.TreePathInterp import TreePathInterp
from argrelay.runtime_context.InterpContext import InterpContext


class FirstArgInterp(TreePathInterp):
    """
    Dispatch command line interpretation to the next interpreter based on the command_id.

    `FirstArgInterp` was re-implemented in terms of `TreePathInterp` (FS_01_89_09_24).

    Implements FS_42_76_93_51.
    """

    def __init__(
        self,
        interp_factory_id: str,
        config_dict: dict,
        interp_ctx: InterpContext,
    ):
        super().__init__(
            interp_factory_id,
            config_dict,
            interp_ctx,
        )



from __future__ import annotations

from dataclasses import dataclass, field


@dataclass
class QueryResult:
    """
    Results of a query via `QueryEngine.query_prop_values`.
    """

    data_envelope: dict = field()
    """
    Last `data_envelope` in search results

    Normally, it is only useful when `found_count` = 1
    """

    found_count: int = field()
    """
    Total number of `data_envelope`-s found.
    """

    remaining_types_to_values: dict[str, list[str]] = field()
    """
    See `EnvelopeContainer.remaining_types_to_values`.
    """



"""
"""

goto_host_funct_ = "goto_host"
goto_service_funct_ = "goto_service"
goto_repo_func_ = "goto_repo"

list_host_func_ = "list_host"
list_service_func_ = "list_service"

desc_commit_func_ = "desc_commit"



from marshmallow import Schema, fields, RAISE, post_load, validates_schema, ValidationError

from argrelay.misc_helper import get_config_path
from argrelay.misc_helper.TypeDesc import TypeDesc
from argrelay.runtime_data.ServerConfig import ServerConfig
from argrelay.schema_config_core_client.ConnectionConfigSchema import connection_config_desc
from argrelay.schema_config_core_server.GuiBannerConfigSchema import gui_banner_config_desc
from argrelay.schema_config_core_server.MongoConfigSchema import mongo_config_desc
from argrelay.schema_config_core_server.QueryCacheConfigSchema import query_cache_config_desc
from argrelay.schema_config_core_server.StaticDataSchema import static_data_desc
from argrelay.schema_config_plugin.PluginEntrySchema import plugin_entry_desc

connection_config_ = "connection_config"
mongo_config_ = "mongo_config"
query_cache_config_ = "query_cache_config"
gui_banner_config_ = "gui_banner_config"
plugin_instance_id_load_list_ = "plugin_instance_id_load_list"
plugin_dict_ = "plugin_dict"
static_data_ = "static_data"


class ServerConfigSchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    connection_config = fields.Nested(
        connection_config_desc.dict_schema,
        required = True,
    )

    mongo_config = fields.Nested(
        mongo_config_desc.dict_schema,
        required = True,
    )

    query_cache_config = fields.Nested(
        query_cache_config_desc.dict_schema,
        required = True,
    )

    gui_banner_config = fields.Nested(
        gui_banner_config_desc.dict_schema,
        required = True,
    )

    # Lists `plugin_instance_id`-s to specify load order for the plugins and their activation.
    plugin_instance_id_load_list = fields.List(
        fields.String(),
        required = True,
    )

    # Plugin config data: key = `plugin_instance_id`, value = `plugin_entry`:
    plugin_dict = fields.Dict(
        keys = fields.String(),
        values = fields.Nested(plugin_entry_desc.dict_schema),
        required = True,
    )

    static_data = fields.Nested(
        static_data_desc.dict_schema,
        required = True,
    )

    # No interp_factories field for schema -
    # it is generated for ServerConfig internally based on `plugin_instance_id_load_list`.

    @post_load
    def make_object(self, input_dict, **kwargs):
        # Populate `plugin_instance_id` from `plugin_dict` into each
        return ServerConfig(
            connection_config = input_dict[connection_config_],
            mongo_config = input_dict[mongo_config_],
            query_cache_config = input_dict[query_cache_config_],
            gui_banner_config = input_dict[gui_banner_config_],
            plugin_instance_id_load_list = input_dict[plugin_instance_id_load_list_],
            plugin_dict = input_dict[plugin_dict_],
            static_data = input_dict[static_data_],
        )

    @validates_schema
    def validate_known(self, input_dict, **kwargs):
        for plugin_instance_id in input_dict[plugin_instance_id_load_list_]:
            # Ensure every `plugin_instance_id` is defined in `plugin_dict`:
            if plugin_instance_id not in input_dict[plugin_dict_]:
                raise ValidationError(f"`{plugin_instance_id}` is not defined in `{plugin_dict_}`")


server_config_desc = TypeDesc(
    dict_schema = ServerConfigSchema(),
    ref_name = ServerConfigSchema.__name__,
    dict_example = {
        connection_config_: connection_config_desc.dict_example,
        mongo_config_: mongo_config_desc.dict_example,
        query_cache_config_: query_cache_config_desc.dict_example,
        gui_banner_config_: gui_banner_config_desc.dict_example,
        plugin_instance_id_load_list_: [
            "some_plugin_instance_id",
        ],
        plugin_dict_: {
            "some_plugin_instance_id": plugin_entry_desc.dict_example,
        },
        static_data_: static_data_desc.dict_example,
    },
    default_file_path = get_config_path("argrelay.server.yaml"),
)



from argrelay.enum_desc.InterpStep import InterpStep
from argrelay.enum_desc.TokenType import get_token_type, TokenType

from argrelay.runtime_context.InterpContext import InterpContext


class AbstractInterp:
    """
    Interpret command line sharing current state via :class:`InterpContext`.

    New instance of interpreter is created by (plugin implementing) `AbstractInterpFactory` for each request.
    """

    def __init__(
        self,
        interp_factory_id: str,
        config_dict: dict,
        interp_ctx: InterpContext,
    ):
        self.interp_factory_id: str = interp_factory_id
        self.config_dict: dict = config_dict
        self.interp_ctx: InterpContext = interp_ctx
        self.base_envelope_ipos: int = interp_ctx.curr_container_ipos

    def __repr__(self) -> str:
        return f"fid: {self.interp_factory_id}, {super().__repr__()}"

    def consume_key_args(self) -> None:
        pass

    def consume_pos_args(self) -> None:
        pass

    def try_iterate(self) -> InterpStep:
        pass

    def run_fill_control(self) -> None:
        pass

    def propose_arg_completion(self) -> None:
        pass

    def next_interp(self) -> "AbstractInterp":
        """
        Return next interp factory id (or None).
        """
        return None

    def is_pos_arg(self, token_ipos: int) -> bool:
        return get_token_type(self.interp_ctx.parsed_ctx.all_tokens, token_ipos) == TokenType.PosArg



from argrelay.relay_client.AbstractClient import AbstractClient
from argrelay.relay_client.RemoteClientCommandFactory import RemoteClientCommandFactory
from argrelay.runtime_data.ClientConfig import ClientConfig


class RemoteClient(AbstractClient):
    """
    Client remote to server

    It talks to `CustomFlaskApp` which is remote server in this case.

    See also `LocalClient` and `LocalServer`.
    """

    def __init__(self, client_config: ClientConfig):
        super().__init__(
            client_config,
            RemoteClientCommandFactory(client_config),
        )
        assert not client_config.use_local_requests



from argrelay.plugin_interp.AbstractInterpFactory import AbstractInterpFactory
from argrelay.plugin_interp.TreePathInterp import TreePathInterp
from argrelay.plugin_interp.TreePathInterpFactoryConfigSchema import tree_path_interp_factory_config_desc
from argrelay.runtime_context.InterpContext import InterpContext


class TreePathInterpFactory(AbstractInterpFactory):
    """
    Implements FS_01_89_09_24.
    """

    def __init__(
        self,
        plugin_instance_id: str,
        config_dict: dict,
    ):
        super().__init__(
            plugin_instance_id,
            config_dict,
        )
        tree_path_interp_factory_config_desc.validate_dict(config_dict)

    def create_interp(
        self,
        interp_ctx: InterpContext,
    ) -> TreePathInterp:
        return TreePathInterp(
            self.plugin_instance_id,
            self.config_dict,
            interp_ctx,
        )



from argrelay.misc_helper.ElapsedTime import ElapsedTime
from argrelay.relay_client import AbstractClientCommand
from argrelay.relay_client.AbstractClientCommandFactory import AbstractClientCommandFactory
from argrelay.runtime_context.InputContext import InputContext
from argrelay.runtime_data.ClientConfig import ClientConfig


class AbstractClient:
    """
    An abstract `relay_client` to `relay_server`
    """

    def __init__(
        self,
        client_config: ClientConfig,
        command_factory: AbstractClientCommandFactory
    ):
        self.client_config: ClientConfig = client_config
        self.command_factory: AbstractClientCommandFactory = command_factory

    # noinspection PyMethodMayBeStatic
    def make_request(self, input_ctx: InputContext) -> AbstractClientCommand:
        input_ctx.print_debug()
        command_obj = self.command_factory.create_command(input_ctx)
        command_obj.execute_command(input_ctx)
        ElapsedTime.measure("after_execute_command")
        return command_obj



from argrelay.handler_response.AbstractClientResponseHandler import AbstractClientResponseHandler
from argrelay.runtime_context.InputContext import InputContext


class AbstractClientCommand:

    def __init__(
        self,
        response_handler: AbstractClientResponseHandler,
    ):
        self.response_handler: AbstractClientResponseHandler = response_handler

    def execute_command(self, input_ctx: InputContext):
        pass



from argrelay.enum_desc.CompType import CompType

from argrelay.handler_request.AbstractServerRequestHandler import AbstractServerRequestHandler
from argrelay.misc_helper.ElapsedTime import ElapsedTime
from argrelay.relay_server.LocalServer import LocalServer
from argrelay.runtime_context.InputContext import InputContext


class ProposeArgValuesServerRequestHandler(AbstractServerRequestHandler):

    def __init__(
        self,
        local_server: LocalServer,
    ):
        super().__init__(
            local_server = local_server,
        )

    def handle_request(self, input_ctx: InputContext) -> dict:
        assert input_ctx.comp_type != CompType.DescribeArgs
        assert input_ctx.comp_type != CompType.InvokeAction

        self.interpret_command(self.local_server, input_ctx)
        ElapsedTime.measure("after_interpret_command")

        return {
            "arg_values": self.interp_ctx.propose_arg_values()
        }



import subprocess
from subprocess import TimeoutExpired, Popen

from argrelay.misc_helper import eprint
from argrelay.mongo_data.MongoConfig import MongoConfig


class MongoServerWrapper:

    def __init__(self):
        self.is_mongomock: bool
        self.is_started: bool = False
        self.mongo_proc: Popen

    def start_mongo_server(self, mongo_config: MongoConfig):

        if mongo_config.use_mongomock_only:
            self.is_mongomock = True
            return

        if mongo_config.mongo_server.start_server and not self.is_started:
            self.mongo_proc = subprocess.Popen(mongo_config.mongo_server.server_start_command, shell = True)
            try:
                self.mongo_proc.wait(timeout = 5)
            except TimeoutExpired:
                # Ignore: it probably started without issues. Or did it?
                pass
            ret_code = self.mongo_proc.poll()
            eprint("mongo_proc: ret_code: ", ret_code)
            if not ret_code:
                # It must be still running:
                self.is_started = True
            elif ret_code != 0:
                # Something went wrong:
                raise RuntimeError

    def stop_mongo_server(self):

        if self.is_mongomock:
            return

        if self.is_started:
            self.mongo_proc.kill()
        while self.is_started:
            try:
                self.mongo_proc.wait(timeout = 5)
            except TimeoutExpired:
                eprint("still running: mongo_proc: ", self.mongo_proc)
            ret_code = self.mongo_proc.poll()
            eprint("mongo_proc: ret_code: ", ret_code)
            if ret_code:
                # Process finished:
                self.is_started = False



from marshmallow import Schema, fields, validates_schema, INCLUDE

from argrelay.enum_desc.ReservedArgType import ReservedArgType
from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.misc_helper.TypeDesc import TypeDesc
from argrelay.schema_config_interp.FunctionEnvelopeInstanceDataSchema import function_envelope_instance_data_desc

mongo_id_ = "_id"
"""
Field used by Mongo DB to uniquely identify every object inside a collection.
If not provided, it is auto-generated as `bson.objectid.ObjectId`.
Note that it is not serializable by `Schema.dump` and must be excluded.
for `argrelay`, if unique id is required, `envelope_id_` must be used (which is copied to `_id` on load).
"""

envelope_id_ = "envelope_id"
"""
Not required (yet) field within `envelope_metadata` with unique id for `data_envelope`.
If provided, it is given to MongoDB as `id_` (otherwise, if not provided, MongoDB auto-generates one).
"""

instance_data_ = "instance_data"
"""
Data specific to `envelope_class`.
Unlike `envelope_payload` `argrelay` does not inspect, `instance_data` can be inspected
(if not inspected by `argrelay`, but by its plugins) and this data has schema implied or defined somewhere.
"""

envelope_payload_ = "envelope_payload"
"""
Data `argrelay` does not inspect.
"""


class DataEnvelopeSchema(Schema):
    """
    Schema for all :class:`StaticDataSchema.data_envelopes`

    Note that this schema definition (unlike many others) is not used to `Schema.dump` `dict` instances
    because `data_envelope`-s contain arbitrary top-level keys used as (search) metadata.
    Because these top-level keys are arbitrary, they cannot be defined in this schema.
    Because they cannot be defined in the schema, they do not survive `Schema.dump`.
    This is a known issue/limitation of `marshmallow` - the `Meta.unknown` field is only used on `Schema.load`
    to allow extra keys in, but `Schema.dump` simply do not serialize them.
    See these fields (which should otherwise use `DataEnvelopeSchema`):
    *   `InvocationInputSchema.data_envelopes`
    *   `EnvelopeContainerSchema.data_envelope`
    *   `StaticDataSchema.data_envelopes`
    """

    class Meta:
        # All other fields of data envelope becomes its metadata available for search queries.
        # Note that it does not work for `Schema.dump`:
        unknown = INCLUDE
        strict = True

    envelope_id = fields.String(
        # TODO: make it required for predictability - isn't it required?
        required = False,
    )

    """
    Data specific to `envelope_class`.
    Each envelope class may define its own schema for that data.
    For example, `ReservedEnvelopeClass.ClassFunction` defines `FunctionEnvelopeInstanceDataSchema`.
    """
    instance_data = fields.Dict(
        # TODO: make it required for predictability - isn't it required?
        required = False,
    )

    """
    Arbitrary schemaless data (payload) wrapped by `DataEnvelopeSchema`.
    It is not inspected by `argrelay`.
    """
    envelope_payload = fields.Dict(
        # TODO: make it required for predictability - isn't it required?
        required = False,
    )

    @validates_schema
    def validate_known(self, input_dict, **kwargs):
        if input_dict[ReservedArgType.EnvelopeClass.name] == ReservedEnvelopeClass.ClassFunction.name:
            function_envelope_instance_data_desc.validate_dict(input_dict[instance_data_])


data_envelope_desc = TypeDesc(
    dict_schema = DataEnvelopeSchema(),
    ref_name = DataEnvelopeSchema.__name__,
    dict_example = {
        envelope_id_: "some_unique_id",
        instance_data_: function_envelope_instance_data_desc.dict_example,
        envelope_payload_: {},
        ReservedArgType.EnvelopeClass.name: ReservedEnvelopeClass.ClassFunction.name,
        "SomeTypeA": "A_value_1",
        "SomeTypeB": "B_value_1",
        "SomeTypeC": "C_value_1",
    },
    default_file_path = "",
)
